"""
analysis scripts written by Gergo
contact: gt2253@cumc.columbia.edu
5/9/2022
5/26/2022 -- moved to lab3/experiment/"""

from dataclasses import dataclass

from os.path import join
import collections

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
# import dabest  # need to install this dependency!


@dataclass

class sleep:
    pass

    # imaging: str
    # eeg: str
    # eeg_folder: im_exp.sima_path+'/eeg'


def calc_ci(x, ci=0.95, bootstrap=1000, decimals=5):
    """calculates the 95 percent confidence interval of an array

    Parameters:
    ===========
    x: numpy.array
        or something in array format
    ci: float, optional
        confidence interval to calcualte
    bootstrap: int, optional
        number of bootstrap resampling
    round: int, optional
        how many decimals should be rounded
        
    Returns:
    ========
    tuple 
        upper and lower confidence interval values rounded to 4 digits"""
    
    confidence = ci
    
    values = [np.random.choice(x,size=len(x),
                               replace=True).mean() for i in range(bootstrap)]
    up, low = np.percentile(values, [100*(1-confidence)/2,
                       100*(1-(1-confidence)/2)])
    
    return up.round(decimals), low.round(decimals)

def ci_difference(awake_arr, nrem_arr, resamples=5000):
    """ 
    uses estimation stats to evaluate the difference between two sets
    of numbers. The documentation can be found here:
    https://acclab.github.io/DABEST-python-docs/tutorial.html
    
    Parameters:
    ===========
    awake_arr: numpy array
        compied dfof of awake periods
    nrem_arr: numpy array
        compiled dfof of NREM periods
    resamples: int
        number of resamples to generate the effect size bootstrap
        
    Returns:
    ========
    tuple
    difference: float
        this is the Cohen D effect size between the two arrays
    significance: float
        the p value for the (unpaired) student test 
    """

    dabest_dict = {'awake': awake_arr, 'nrem': nrem_arr}
    dabest_df= pd.DataFrame.from_dict(dabest_dict, orient='index')
    dabest_df = dabest_df.transpose()

    # calculating the stats
    two_groups_unpaired = dabest.load(dabest_df, idx=("awake", "nrem"),
                                      resamples=resamples)
    stats = two_groups_unpaired.cohens_d
    significance =stats.results['pvalue_students_t'][0] < 0.05
    difference = stats.results['difference'][0]

    return (difference, significance)


def converter(series, state):
    """converts a series to a dataframe. This was used for creating a
    boolean column for brain states 

    Parameters:
    ===========
    series: pandas.Series
    state: list 
    """
    df = series.reset_index()
    df.set_index(['cells'], inplace=True)
    df.rename(columns={'dfof': 'dfof_'+state}, inplace=True)
    result_df = df[df[state]==True]    
    return result_df.drop(labels=state, axis=1)

def upsample(df, freq):
    """
    Upsampling dataframes, alternative for pd.resample method which results in less rows.
    If you resample you don't change the boundaries. You still have 4 initial values, 
    and you add as many values as there are intermediates (4 values -> 3 intermediates, 
    n values -> n-1 intermediates)

    Parameters:
    ===========
    df: pandas DataFrame with a Timedelta index or timedelta column
    freq: str
        the upsampling frequecy. E.g. '0.5S', '0.05S'

    Returns:
    ========
    upsampled dataframe

    Note: the `closed` arguent has changed to `inclusive` in pandas >=1.4
    """

    delta = df.index[1]-df.index[0]
    idx = pd.date_range(df.index.min(),
                        df.index.max()+delta,
                        closed='left',
                        freq=freq)

    return df.reindex(idx).ffill()

def sub(a,b):
    """
    Subtracts two tuples from each other
    Params:
    =======
    a, b: tuple 
    Returns:
    ========
    tuple
    """
    return tuple(map(lambda x, y: x-y, a,b))

def overlap_test(tuple1, tuple2):
    """ 
    Checks whether the items in tuple2 are within the range of tuple1. Used to
    check whether confidece intervals overlap.
    
    Paramaters:
    ===========
    tuple1, tuple2: tuple, list
        with 2 items in each
    Return:
    =======
    True: if the item1 or item2 is between the numbers of tuple1
    False: if the item1 or item2 is NOT between the numbers of tuple1
    """
    return ((tuple1[0]<= tuple2[0] <=tuple1[1]) or
            (tuple1[0]<= tuple2[1] <=tuple1[1]))

def transient_freq_count(df, state, state_frames):
    """
    counts transient frequency. 

    df: pandas DataFrame
        must have 'roi_labels','onset' and 'state: ' columns
        >>> expt_transients = im_exp.events(signal_type='imaging',
                                event_type='transients',
                                channel='Ch2',
                                label='suite2p')
        >>> trans = expt_transients.dataframe

    state: str
        usually it's in 'state:NREM' format
    state_frames: int
        number of frames for the given state

    Return:
    =======
    freq_df: pandas DataFrame
        
    """
    freq_df = df[df[state]].groupby('roi_label')['onset'].count().reset_index()
    freq_df['frames'] = state_frames
    freq_df['State'] = state
    freq_df['Freq (event/sec)'] = state_frames
    freq_df['Freq (event/sec)'] = freq_df['onset'] / freq_df['frames']
    return freq_df

def import_eeg(imaging_exp, eeg_file):
    """
    imports scored eeg data from a csv file. The file should be located
    in a subfolder named 'eeg' within the sima folder
    
    Parameters:
    ===========
    imaging_exp: ImagingExperiment object
    eeg_file: str
        imagaging file name
    Return:
    ======
    eeg_df: pandas DataFrame
    
    """
    eeg_data_path = imaging_exp.sima_path+"/eeg/"+eeg_file
    eeg_df = pd.read_csv(eeg_data_path, names=['time', 'eeg'])
    eeg_df['eeg'] = eeg_df['eeg'].astype(int)
    eeg_df['awake'] = eeg_df['eeg'] == 0
    eeg_df['NREM'] = eeg_df['eeg'] == 1
    eeg_df['REM'] = eeg_df['eeg'] == 2
    return eeg_df

def plot_state_velocity(imaging_exp, df, state):
    """
    Creates a graphical representation of velocity and brain state

    Parameters:
    ===========
    imaging_exp: ImagingExperiment object
    df: pandas DataFrame
        required columns: 'filtered velo', 'NREM','state:awake',
        'state:locomotion'
    state: list, str
        list or str of state columns from the df with booleans or 0-1 values
    """
    
    fig, ax1 = plt.subplots(nrows=2, ncols=1, figsize=(15,10))
    fig.suptitle('Velocity vs %s' % state, fontsize=16, y=1.02)

    color = 'tab:red'
    ax1[0].set_xlabel('frames')
    ax1[0].set_ylabel('m/s', color=color, fontsize=16)
    df['filtered velo'][:18000].plot(ax=ax1[0], color=color)
    ax1[0].tick_params(axis='y', labelcolor=color)
    ax1[0].legend().remove()

    ax2 = ax1[0].twinx()  # instantiate a second axes that shares the same x-axis

    color = 'tab:blue'
    ax2.set_ylabel(state, color=color, fontsize=16)    
    df[state].astype(int)[:18000].plot(ax=ax2,color=color)
    ax2.set_yticks((0,1))
    ax2.tick_params(axis='y', labelcolor=color)
    ax2.legend().remove()

    color = 'tab:red'
    ax1[1].set_xlabel('frames')
    ax1[1].set_ylabel('m/s', color=color, fontsize=16)
    df['filtered velo'][18000:].plot(ax=ax1[1], color=color)
    ax1[1].tick_params(axis='y', labelcolor=color)
    ax1[1].legend().remove()

    ax2 = ax1[1].twinx()  # instantiate a second axes that shares the same x-axis

    color = 'tab:blue'
    ax2.set_ylabel(state, color=color, fontsize=16)  
    df[state][18000:].astype(int).plot(ax=ax2,color=color)
    ax2.set_yticks((0,1))
    ax2.tick_params(axis='y', labelcolor=color)
    ax2.legend().remove()

    fig.tight_layout()  # otherwise the right y-label is slightly clipped
    fname = state.replace(':', '_')
    fig.savefig(join(imaging_exp.sima_path, 'Velocity vs %s.png' % fname),
               transparent=True, dpi=300, bbox_inches='tight')

def brain_state_filter(velo_eeg_df, states):
    """
    sets up boolean masks for brain states

    Paramaters:
    ===========
    velo_eeg_df: pandas DataFrame
        contains a 'filtered_velo' columns with the velocity and columns
        with brain states. See eeg.import_eeg()
    states: list
        list of the possible brain states: NREM, REM, awake
        
    Return:
    =======
    filters: dictionary
        the keys correspond to boolean values
    states: list
        the states used as an input plus 'locomotion'

    """
    l1 = ['NREM', 'REM', 'awake']
    filters = {}
    if collections.Counter(states) == collections.Counter(l1):
        print(f'Making filters for {l1} and locomotion')        
        awake = ((velo_eeg_df['NREM']==False) &
             (velo_eeg_df['REM']==False) &
             (velo_eeg_df['filtered velo']<0.1))
        NREM = ((velo_eeg_df['NREM']) &
                (velo_eeg_df['filtered velo']<=0.1))
        REM = ((velo_eeg_df['REM']) &
               (velo_eeg_df['filtered velo']<=0.1))
        locomotion = ((velo_eeg_df['NREM']==False) &
              (velo_eeg_df['REM']==False) &
              (velo_eeg_df['filtered velo']>0.1))
        filters = {'awake':awake,
                  'NREM': NREM,
                  'REM':REM,
                  'locomotion':locomotion}        
    else:
        print(f'no REM, making filters for NREM, awake and locomotion')
        awake = ((velo_eeg_df['NREM']==False) &
             (velo_eeg_df['REM']==False) &
             (velo_eeg_df['filtered velo']<0.1))
        NREM = ((velo_eeg_df['NREM']) &
                (velo_eeg_df['filtered velo']<=0.1))        
        locomotion = ((velo_eeg_df['NREM']==False) &
              (velo_eeg_df['REM']==False) &
              (velo_eeg_df['filtered velo']>0.1))
        filters = {'awake': awake,
                  'nrem':nrem,
                  'locomotion':locomotion}
    states.append('locomotion')
    return filters, states


from __future__ import annotations
import pandas as pd
import numpy as np
from functools import reduce
from typing import TYPE_CHECKING

from lab3.core.filters import Filter, TimeFilter, _BoundFilter
from lab3.event import SpatialEvent

from .time_filters import RealTimeEventFilter

class JointFilter(Filter):
    """Implements more complex filtering operations that depend on rows and columns
    e.g. 
    """
    pass

class _BoundJointFilter(_BoundFilter):
    def __call__(self, dataframe):
        return dataframe

class _JointFilterIdentity(Filter):
    def bind_to(self, expt):
        return _BoundJointFilter(index=None, columns=None, level=None)

JointFilterIdentity = _JointFilterIdentity()

class GroupbyFilterMixin(JointFilter):
    def __init__(self, *args, groupby=None, **kwargs):
        self.groupby = groupby
        super().__init__(*args, **kwargs)

    def bind_groupby(self, bound_filters):
        return _BoundGroupbyFilter(bound_filters, groupby=self.groupby)

class ROIEventFilter(GroupbyFilterMixin, RealTimeEventFilter):
    constructor_ids = ['event_spec', 'pre', 'post', 'groupby']

    def bind_to(self, expt):

        events = expt.events(**self.event_spec)
        imaging_times = expt.imaging_times()

        bound_filters = {}

        for roi_events in events:
            idxes = []
            for ev in roi_events:
                lower, = np.where(imaging_times >= ev.onset - self.pre)
                upper, = np.where(imaging_times <= ev.offset + self.post)

                idxes.append(pd.RangeIndex(lower[0], upper[-1]))

            bound_filters[roi_events.name] = self.bind(reduce(pd.Int64Index.union, idxes, pd.Int64Index([])))

        return self.bind_groupby(bound_filters)

class _BoundGroupbyFilter(_BoundJointFilter):

    constructor_ids = []

    def __init__(self, bound_filters, groupby):
        self.bound_filters = bound_filters
        self.groupby = groupby


    def __call__(self, dataframe):
        key = dataframe.level[self.groupby].unique()
        #assert len(key) == 1
        #key = key[0]

        for k in key:
            dataframe.loc[dataframe.level[self.groupby] == k, ~self.bound_filters[k]] = np.nan

        return dataframe




"""Convenience function wrappers for experiment methods"""

import inspect
from functools import wraps

from sima.misc import most_recent_key


def get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def default_label_handling(func):
    """Decorator for selecting a default ROI label if None is passed. Defaults
    to the roi label for the most recently created ROIList in the imaging
    dataset

    The wrapped function should take an ImagingMixin-derived object as an
    argument and a label as a keyword argument.
    """

    @wraps(func)
    def set_default_label(*args, **kwds):

        kwargs = get_default_args(func)
        kwargs.update(kwds)

        if kwargs['label'] is None:
            kwargs['label'] = most_recent_key(args[0].imaging_dataset.ROIs)
            print(f'Label not passed. Using label {kwargs["label"]}')

        return func(*args, **kwargs)

    return set_default_label



from os.path import join

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.backends.backend_pdf import PdfPages

import numpy as np
import pandas as pd
import seaborn as sns

from scipy.stats import stats
from statsmodels.distributions.empirical_distribution import ECDF


from lab3.plotting import utils 

def prep_csv(im_exp, f_name, state, direction="Upregulated"):
    """loads a csv file and returns a list of the significant cells
    
    Parameters:
    ===========
    im_exp: lab3 ImagingExperiment object
    f_name: str
        name of the csv file to load
    state: str
        brain state associated with the csv as it appears in 
        the CSV file
	direction: str, default="Upregulated"
		'Upregulated' will pull the upregulated cells,
		'Downregulated is the downregulated ones

    Return:
    =======
    ind_list: list
    	list of ROI labels

    """
    sig_cells = pd.read_csv(join(im_exp.sima_path, f_name))
    sig_cells.drop(labels='Unnamed: 0', inplace=True, axis=1)
    sig_cells.rename({'cells': 'roi_label'}, inplace=True, axis=1)
    sig_cells.set_index('roi_label', inplace=True)
    ind_list = list(sig_cells[
        (sig_cells['Direction']==direction) &
                             (sig_cells['State']=='dfof '+state)].index)
    return ind_list

def xcorr_comparison(im_exp, df1, df2, st1, st2, dtype,
					lim=None, save=False, **kwargs):
	"""
	compares two correlation matrices and uses

	Parameters:
	===========
	im_exp: lab3 ImagingExperiment object
	df1, df2: pandas DataFrame
		contains the data to be correlated
	st1, st2: string
		brain states for df1 and df2 respectively
	dtype: list of 2 str
		needed for plotting and saving. Type of the data
		being plotted (Calcium or spike), up or downregulated
		cells (up or down)
	lim: tuple of integers (max, min), optional
		the upper and lower limits of the color coding. 
	save: boolean

	Returns:
	========
	fig: pyplot figure object 
		with 4 subplots

	"""

	state1 = utils.upper(df1.corr())
	state2 = utils.upper(df2.corr())
	ecdf_state1= ECDF(state1)
	ecdf_state2 = ECDF(state2)
	if lim == None:
		vmax = np.array([state1.max(), state2.max()]).max()
		vmin = np.array([state1.min(), state2.min()]).min()		
	else:
		vmax, vmin =lim

	fig, axs = plt.subplots(ncols=2, nrows=2, figsize=(10,8))

	utils.corr_plot(df1, title=st1, ax=axs[0][0],
	         xticklabels=False, yticklabels=False, cbar=False, **kwargs)
	utils.corr_plot(df2, title=st2, ax=axs[0][1],
	         xticklabels=False, yticklabels=False, **kwargs)

	axs[1][0].plot(ecdf_state1.x, ecdf_state1.y, label=st1)
	axs[1][0].plot(ecdf_state2.x, ecdf_state2.y, label=st2)
	axs[1][0].spines['top'].set_visible(False)
	axs[1][0].spines['right'].set_visible(False)
	axs[1][0].legend()
	axs[1][0].set_title('Cumulative Distribution of pearson r-s')

	significance_test(df1.corr(), df2.corr(), ax=axs[1][1])
	fig.suptitle(f'{dtype[0]} activity cross correlation of significanlty ' \
		f'{dtype[1]}regulated cells')

	if save:
		fig.savefig(join(im_exp.sima_path,
	    			f'{dtype[0]}_xcorr_w_CDF_{st1}_vs_{st2}'\
							f'of_significantly_{dtype[1]}regulated_cells.png'),
	           dpi=300, transparent=False, tight_layout=True)
	return fig

def significance_test(m1, m2, ax=None):
	"""Nonparametric permutation testing Monte Carlo. Used to compare
	the similarities of two distribution obtained with cross correlation.

	Parameters:
	==========
	m1,m2: pandas or numpy correlation matrix
	ax: pyplot ax object

	Return:
	=======


	"""
	np.random.seed(0)
	rhos = []
	n_iter = 5000
	true_rho, _ = stats.spearmanr(utils.upper(m1), utils.upper(m2))
	# matrix permutation, shuffle the groups
	m_ids = list(m1.columns)
	m2_v = utils.upper(m2)
	for iter in range(n_iter):
		np.random.shuffle(m_ids) # shuffle list
		r, _ = stats.spearmanr(utils.upper(m1.loc[m_ids, m_ids]), m2_v)
		rhos.append(r)
	perm_p = ((np.sum(np.abs(true_rho) <= np.abs(rhos)))+1)/(n_iter+1) # two-tailed test

	if ax==None:
		f,ax = plt.subplots()
	else:
		ax=ax
	ax.hist(rhos, bins='auto')
	ax.axvline(true_rho, color = 'r', linestyle='--')
	ax.set(title=f"Spearman r: {round(true_rho, 2)}, Permuted p: {perm_p:.3f}",
	       ylabel="counts", xlabel="rho")

	return ax

def xcorr_lag(datax, datay, lag=0):
	""" Lag-N cross correlation. 

	Parameters
	==========
	datax, datay : pandas.Series objects of equal length
	lag : int, default 0    

	Returns
	=======
	crosscorr : float
	"""

	return datax.corr(datay.shift(lag))


import os
import numpy as np
import scipy.sparse as sparse
import h5py
import warnings

from packaging import version

# THIS PACKAGE IS DEPRECATED! 
# My environment does not support h5py >= 2.9 so I use this to control
# caching, but use h5py native cache if possible. -Z
H5_VERSION = h5py.__version__
if version.parse(H5_VERSION) < version.parse("2.9"):
    warnings.warn(f"h5py version {H5_VERSION} detected, "
                f"using deprecated `h5py_cache` package! "
                "(upgrade to h5py >= 2.9 to bypass this warning)")
    import h5py_cache

import sima
from sima import ImagingDataset
from sima.sequence import _Sequence_HDF5

from ...experiment.base import ImagingOnlyExperiment
from ...misc.sima_helpers import backup_dir
from . import trefide_denoiser 
from . import convertToMemmap

FLOATSIZE = 8 # bytes
H5_CHUNKSIZE = 10*10**6 # bytes
N_CHUNKS = 10

FASTSCRATCH = "/fastscratch/memmaps"

def import_denoised_signals(expt):
    denoised_ds = DenoisedDataset.load(expt.sima_path)
    denoised_expt = ImagingOnlyExperiment(denoised_ds.denoised_savedir)

    with denoised_expt.signals_file(mode='r') as origin, \
        expt.signals_file(mode='a') as target:

        for k in origin.keys():
            print(f'Importing {k}...')
            target.put(k, origin[k])

    print('Importing ROIs...')
    expt.imaging_dataset.add_ROIs(denoised_expt.roi_list(label='denoised'), 'denoised')

class _DenoisedSequence(_Sequence_HDF5):
    """This works in a FUNDAMENTALLY DIFFERENT way from other sima 
    sequences. Rather than storing every frame, this stores the spatial 
    and temporal information of the session separately (similar to an SVD)
    and reconstitutes them on the fly as needed.

    This allows compression to occur, as biologically-relevant signals
    occupy a much lower dimensional manifold in pixel space than 2**18 (=512 x 512)

    If Y_yxt is a tensor representing the movie, then we have
        Y_yxt = U_yxk.V^k_t
    where k is the latent dimension, with log10(k) ~ 2. 
    U and V are stored instead of Y.

    (Don't put @sima_compatible here, it's not going to work.)
    """
    def __init__(self, path, group='/denoised0'):
        self._path = os.path.abspath(path)
        if version.parse(H5_VERSION) >= version.parse("2.9"):
            # Use h5py-native caching (>= 2.9 only)
            self._file = h5py.File(path, 'r', 
                                    rdcc_nbytes=N_CHUNKS*H5_CHUNKSIZE, # cache optimizations for performance
                                    rdcc_nslots=100*N_CHUNKS)
        else:
            # Use deprecated h5py_cache package
            self._file = h5py_cache.File(path, 'r', 
                            chunk_cache_mem_size=N_CHUNKS*H5_CHUNKSIZE) #  # h5py

        self._group = self._file[group]
        # self._spatial_shape = self._group['U'].shape[:-1]
        # self._nk = self._group['U'].shape[-1]

        # This is needed for every frame so just load it into RAM
        U = np.array(self._group['U'])
        self._spatial_shape = U.shape[:-1]
        self._nk = U.shape[-1]
        resh = U.reshape(np.product(self._spatial_shape), self._nk)
        self._spatial = sparse.csr_matrix(resh)
        # self._spatial = sparse.csr_matrix(np.array(self._group['U']).reshape( 
        #                     (np.product(self._spatial_shape), self._nk))) # optimizations... 
        self._temporal = np.array(self._group['V'])
        #self._temporal = np.array(self._group['V'])
        
    def __len__(self):
        return self._temporal.shape[1]
    
    def _get_frame(self, t):
        """This is where the reconstitution happens.
        Shape of U is yxk, shape of V is kt
        """
        flat = self._spatial @ self._temporal[:, t]
        return flat.reshape((1,) + self._spatial_shape + (1,))
    
    def _todict(self, savedir=None):
        d = {'__class__': self.__class__,
             'group': self._group.name}
        if savedir is None:
            d.update({'path': os.path.abspath(self._path)})
        else:
            d.update({'_abspath': os.path.abspath(self._path),
                      '_relpath': os.path.relpath(self._path, savedir)})
        return d

class DenoisedDataset(ImagingDataset):
    """Interface for using Trefide denoiser with SIMA. DenoisedDatasets
    are instantiated via the load method, inherited from ImagingDataset. They
    are not initialized directly. You should only initialize a DenoisedDataset
    from an already motion-corrected dataset.

    Examples
    --------
    >>> from lab3.extraction.denoising import DenoisedDataset
    >>> ds = DenoisedDataset.load("/path/to/my/folder.sima")
    >>> ds.denoise()

    Denoised datasets are saved to a *NEW* h5 and sima folder 
    (/old/path/OLDNAME_denoised.{h5|sima}) as the underlying data 
    is changed.

    This package also implements compression.

    See Buchanan et al. 2019 for details. 
    https://www.biorxiv.org/content/10.1101/334706v4
    """

    def denoise(self, overwrite=False, reprocess=True, on_scratch=True, first_obs=1, 
                signal_channel='0', block_height=40, block_width=40, t_sub=2,  
                n_processes=8, sima_compatible=False, strict_checking=False, **kwargs):
        """Run Trefide denoising on the imaging dataset. It should ordinarily not be 
        necessary to change the Trefide parameters from defaults.

        Parameters
        ----------
        sima_compatible : bool, optional
            Whether to store the denoised HDF5 in full format (i.e., all frames) 
            or in U, V format. The advantage of sima_compatible is you can watch the 
            denoised video on the vis_server. Default False.
        strict_checking : bool, optional
            Whether to validate the data. Default False.
        overlapping : bool, optional
            Whether to overlap blocks the video is split into. Default True.
        block_height : int, optional 
            Frame height will be trimmed to a multiple of this number. Default 40.
        block_width: int, optional
            Frame width will be trimmed to a multiple of this number. Default 40.
        d_sub : int, optional
            Factor by which to downsample spatial dimensions. Default 2. 
        t_sub : int, optional 
            Factor by which to downsample time. Default 2. 
        max_components : int, optional
            Cap on the maximum number of components. Default 50. 
        max_iters_main : int, optional
            Cap on maximum number of iterations in main method. Default 10. 
        max_iters_init : int, optional
            Cap on maximum number of initialization iterations. Default 40. 
        consec_failures : int, optional
            Number of consecutive failures to tolerate. Default 3. 
        tol : float, optional
            Tolerance. Default 5e-3. 
        enable_temporal_denoiser : bool, optional
            Whether to use temporal denoising. Default True. 
        enable_spatial_denoiser : bool, optional
            Whether to use spatial denoising. Default True.

        See also
        --------
        lab3.extraction.denoising.trefide_denoiser.denoise
        """
        
        sequences = []
        
        for i, seq in enumerate(self.sequences):
            try:
                self.validate(seq)
            except AssertionError:
                if strict_checking:
                    raise
                else:
                    pass

            self.mmapname = self.write_memmap(seq, block_height=block_height, 
                                            block_width=block_width, t_sub=t_sub, 
                                            overwrite=overwrite, reprocess=reprocess, 
                                            on_scratch=on_scratch, first_obs=first_obs, 
                                            signal_channel=signal_channel, n_processes=n_processes)

            new_seq = self.denoise_mmap(self.mmapname, i, sima_compatible=sima_compatible,
                                        block_height=block_height, block_width=block_width,
                                        t_sub=t_sub, **kwargs)
            
            sequences.append(new_seq)

        if os.path.exists(self.denoised_savedir):
            backup_dir(self.denoised_savedir, delete_original=True)
            
        if sima_compatible:            
            new_ds = ImagingDataset(sequences, self.denoised_savedir, 
                                channel_names=self.channel_names)
        else:
            new_ds = self.__class__(sequences, self.denoised_savedir, 
                    channel_names=self.channel_names)

        self.cleanup()

    def denoise_mmap(self, mmapname, i, block_height, block_width, t_sub, 
                    sima_compatible=False, **kwargs):
        U, V, mov_denoised = trefide_denoiser.denoise(self.mmapname, block_height=block_height, 
                                                    block_width=block_width, t_sub=t_sub,
                                                    sima_compatible=sima_compatible, **kwargs)
        
        if os.path.exists(self.denoised_h5):
            backup_dir(self.denoised_h5, delete_original=True)

        self.save_denoised_h5(i, U, V, mov_denoised=mov_denoised)
        
        if sima_compatible:
            new_seq = sima.Sequence.create('HDF5', self.denoised_h5, 'yxt', key=f'imaging{i}')
        else:
            new_seq = _DenoisedSequence(self.denoised_h5, group=f"/denoised{i}")

        return new_seq
            
    @property
    def denoised_savedir(self):
        savedir = self.savedir
        if 'denoised' not in savedir:
            # IMPORTANT! *DON'T* write this to a normal sima
            savedir = os.path.splitext(os.path.normpath(savedir))[0] + '_denoised.sima'
        
        return savedir
    
    @property
    def denoised_h5(self):
        
        root, base = os.path.split(self.denoised_savedir)
        base = os.path.splitext(base)[0]
        return os.path.join(root, f'{base}.h5')
        
    def save(self, savedir=None):
        super().save(savedir=self.denoised_savedir)
        
    def save_denoised_h5(self, i, U, V, mov_denoised=None):
        
        with h5py.File(self.denoised_h5, 'a') as f:
            try:
                grp = f[f'denoised{i}']
            except KeyError:
                grp = f.create_group(f"denoised{i}")
            
            grp['U'] = np.pad(U, ((self.bottom, self.top), (self.left, self.right), (0,0)), 
                                constant_values=np.nan)

            V = np.pad(V, ((0,0), (self.beginning, 0)), constant_values=np.nan)

            kdim, tdim = V.shape
            t_chunk_size = int(H5_CHUNKSIZE // (kdim*FLOATSIZE))
            t_chunk_size = min(t_chunk_size, tdim)

            Vdset = grp.create_dataset("V", data=V, 
                                        chunks=(kdim, t_chunk_size))
            
            if mov_denoised is not None:
                try:
                    grp = f[f'/']                
                except KeyError:
                    grp = f.create_group(f"/")
                
                mov_dset = grp.create_dataset(f"imaging{i}", 
                                            shape=U.shape[:-1] + V.shape[1:])

                # TODO: @jit doesn't work with h5py unfortunately
                # EITHER store as memmap (wasteful of hard disk space) or
                # execute as python (slow) or write custom parallel I/O (fiddly)
                # or store U and V and create memmap when needed
                trefide_denoiser.rebuild_movie(U, V, out=mov_dset)
                #grp[f"imaging{i}"] = mov_denoised

    def validate(self, seq):
        data_size = np.product(seq.shape) * FLOATSIZE
        print(f"The whole sequence ({data_size//10**6}MB) will be loaded into RAM")

        assert seq.shape[1] == seq.shape[-1] == 1, \
            "Denoising currently only supports single-plane, single-channel datasets " \
            "(but see the Zequence wrapper if you want to do this)"

        assert seq.shape[0] > seq.shape[1], \
            "Sequence must have at least as many frames as pixels, otherwise this will cause SegFault! "\
            "Do not ignore this warning!"
        
        assert isinstance(seq, sima.sequence._MotionCorrectedSequence), \
            "Is this data motion corrected? Denoising will only give meaningful results " \
            "for motion corrected sequences"

    def write_memmap(self, seq, block_height, block_width, t_sub, 
                    overwrite=False, reprocess=True, on_scratch=True, 
                    first_obs=50, signal_channel='Ch2', n_processes=8):

        test_frame = seq._get_frame(0)
        _, y_valid, _ = np.where(~np.all(np.isnan(test_frame), axis=1))
        ymin, ymax = y_valid[0], y_valid[-1]
        print(ymin, ymax)

        _, x_valid, _ = np.where(~np.all(np.isnan(test_frame), axis=2))
        xmin, xmax = x_valid[0], x_valid[-1]
        print(xmin, xmax)

        seq = seq[:, :, ymin:ymax, xmin:xmax, :]        

        fov_height = (seq.shape[2]//block_height)*block_height
        fov_width = (seq.shape[3]//block_width)*block_width
        fov_depth = 1
        num_frames = (seq.shape[0] // t_sub)*t_sub
        name = os.path.basename(os.path.splitext(os.path.normpath(self.savedir))[0])
        name = name.replace('_', '-') # '_' is significant to CaImAn!

        mmapname = f"{name}_d1_{fov_height}_d2_{fov_width}_d3_{fov_depth}_order_F_frames_{num_frames}_.mmap"

        if on_scratch:
            mmapname = os.path.join(FASTSCRATCH, mmapname)

        # TODO: convert assumes sequence 0
        if not os.path.exists(mmapname) or overwrite:
            btrim = (seq.shape[2] - fov_height)//2    
            ltrim = (seq.shape[3] - fov_width)//2
            ttrim = seq.shape[0] - num_frames

            self.bottom = ymin+btrim
            self.top = test_frame.shape[1] - (fov_height+self.bottom)
            self.left = xmin+ltrim 
            self.right = test_frame.shape[2] - (fov_width+self.left)
            self.beginning = ttrim

            #real_mmap = mmap.transpose((3,2,0,1))[btrim:btrim+fov_height, ltrim:ltrim+fov_width, ttrim:, 0]

            convertToMemmap.convert(seq[ttrim:, :, btrim:btrim+fov_height, ltrim:ltrim+fov_width],
                                    mmapname, n_processes=n_processes, channel=signal_channel, dtype=np.double,
                                    first_obs=first_obs, n_frames=self.num_frames, order='F')
        elif reprocess:
            pass # use existing mmap
        else:
            raise OSError(f"`{self.mmapname}` already exists (reprocess=True to extract from "\
                            f"existing .mmap or overwrite=True to overwrite)")

        return mmapname

    def cleanup(self):
        os.remove(self.mmapname)
        
    def __repr__(self):
        return ('<DenoisedDataset: '
                + f'num_sequences={self.num_sequences},'
                + f'frame_shape={self.frame_shape}, '
                + f'num_frames={self.num_frames}>')


    @property
    def mmapname(self):
        try:
            return self._mmapname
        except AttributeError:
            name = os.path.basename(os.path.splitext(os.path.normpath(self.savedir))[0])
            name = name.replace('_', '-') # '_' is significant to CaImAn!

            z, y, x, c = self.frame_shape
            t = self.num_frames

            mmapname = f"{name}_d1_{y}_d2_{x}_d3_{z}_order_F_frames_{t}_.mmap"
            self._mmapname = mmapname
            return mmapname

    @mmapname.setter 
    def mmapname(self, value):
        self._mmapname = value



"""Common metrics calculate from spatial tuning curves and place fields"""

import numpy as np
import pandas as pd
from astropy.stats import kuiper_two

from lab3.core import Analysis
from .abstract_spatial_tuning import EventSpatialTuning, Occupancy

from lab3.core.helpers import save_load

# Should these generically take single or multiple tuning curves/place fields?
# Probably just single, and use pandas to apply row-wise on the corresponding
# dataframes


class SkaggsSpatialInformation(EventSpatialTuning):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.spatial_tuning_strategy = EventSpatialTuning(*args, **kwargs)
        self.occupancy_strategy = Occupancy(*args, **kwargs)
    
    def apply_to(self, expt, event_spec, **kwargs):
        event_tuning = self.spatial_tuning_strategy.apply_to(expt, event_spec=event_spec, **kwargs)
        occupancy = self.occupancy_strategy.apply_to(expt, signal_spec={}, **kwargs).sum(axis=0)
        
        return self.calculate(event_tuning, occupancy)
    
    
    def calculate(self, event_tuning, occupancy):
        norm_occupancy = occupancy / occupancy.sum()
        
        ratio = event_tuning.div(event_tuning.mean(axis=1), axis=0)
        log2ratio = np.nan_to_num(np.log2(ratio))
        
        return norm_occupancy * ratio * log2ratio

class Kuiper(Analysis):
    def to_columns(self):
        return pd.Index(['statistic', 'p-value'])
    
    @save_load
    def apply_to(self, expt, event_spec, **kwargs):
        events = expt.events(**event_spec)
        index  = expt.signals().index
        
        positions = expt.format_behavior_data()['treadmillPosition']
        imaging_times = expt.imaging_times()
        
        return self.calculate(events, positions, imaging_times, index)

    
    def calculate(self, events, positions, imaging_times, index):
        def _to_position(onset):
            idx = np.searchsorted(imaging_times, onset) - 1
            return positions[idx]
        
        event_positions = events.onset.map(_to_position)
        
        try:
            kuiper_result = kuiper_two(event_positions, positions)
        except:
            kuiper_result = (np.nan, np.nan)
        
        result = pd.DataFrame([kuiper_result], 
                              index=index, columns=self.columns)
        
        return result

def spatial_reliability():
    return np.nan


def spatial_tuning_gain():
    raise NotImplementedError


def place_field_gain():
    raise NotImplementedError


def first_lap():
    """Return the lap number for the first time the cell fired in its place
    field"""
    return np.nan



import os
import pickle as pkl

import numpy as np
import pandas as pd
from scipy.optimize import linear_sum_assignment
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import seaborn as sns
from IPython.display import clear_output

from sima import ImagingDataset
from sima.ROI import ROIList
from lab3.extraction.crossreg import utils
from lab3.signal import SignalFile


class CrossRegistrar:

    """Class for running ROI cross registration across multiple imaging
    datasets.

    Examples
    --------
    >>> # Create a new CrossRegistrar object
    >>> datasets = [ImagingDataset.load(f) for f in sima_dirs]
    >>> cr = CrossRegistrar(datasets, "rois", savedir="/path/to/folder.xreg")
    >>>
    >>> # Run registration with default options
    >>> cr.register_rois()
    >>>
    >>> # Update signal files of original datasets
    >>> cr.update_signals()
    >>>
    >>> # Load an existing CrossRegistrar object
    >>> cr = CrossRegistrar.load("/path/to/existing/folder.xreg")
    >>> cr.register_rois(overwrite=True)
    >>> cr.update_signals(overwrite=True)

    Parameters
    ----------
    datasets : list
        List of sima ImagingDatasets from which to cross register ROIs
    label : str
        Label of ROI lists to cross register
    savedir : str, optional
        Where to save registration progress. Useful if you want to stop and
        resume processing. Defaults to None, which does not save intermediate
        results.

    Attributes
    ----------
    """

    def __init__(self, datasets, label, savedir=None):

        if datasets is None:
            # Special case used to load an existing cross registration object
            if not savedir:
                raise Exception('Cannot initialize without datasets or a'
                                ' directory.')

            with open(os.path.join(savedir, 'info.pkl'), 'rb') as f:
                info = pkl.load(f)
            self._savedir = savedir
            self.label = info['label']
            self.ds_dirs = info['ds_dirs']
            self.datasets = [ImagingDataset.load(ds_dir)
                             for ds_dir in self.ds_dirs]

        elif all(isinstance(ds, ImagingDataset) for ds in datasets):
            self.savedir = savedir
            self.label = label
            self.datasets = datasets
            self.ds_dirs = [ds.savedir for ds in self.datasets]
            if self.savedir is not None:
                self.save()
        else:
            raise TypeError('CrossRegistrar objects must be initialized with '
                            'a list of sima ImagingDatasets.')

    @property
    def num_datasets(self):
        if not hasattr(self, '_num_datasets'):
            self._num_datasets = len(self.datasets)
        return self._num_datasets

    @property
    def datasets(self):
        return self._datasets

    @datasets.setter
    def datasets(self, datasets):
        num_planes = set([ds.frame_shape[0] for ds in datasets])
        assert len(num_planes) == 1, \
            "All datasets must have the same number of planes."
        assert all(self.label in ds.ROIs.keys() for ds in datasets), \
            f"Not all datasets contain ROI label '{self.label}'"
        self._datasets = datasets
        self.num_planes = list(num_planes)[0]

    @property
    def savedir(self):
        return self._savedir

    @savedir.setter
    def savedir(self, savedir):
        if savedir is None:
            self._savedir = None
        elif hasattr(self, '_savedir') and savedir == self.savedir:
            return
        else:
            if hasattr(self, '_savedir'):
                orig_dir = self.savedir
            else:
                orig_dir = False
            savedir = os.path.abspath(savedir)
            if not savedir.endswith('.xreg'):
                savedir += '.xreg'
            os.makedirs(savedir)
            self._savedir = savedir
            if orig_dir:
                from shutil import copy2
                for f in os.listdir(orig_dir):
                    if f.endswith('.pkl') or f.endswith('.h5'):
                        try:
                            copy2(os.path.join(orig_dir, f), self.savedir)
                        except IOError:
                            pass

    def save(self, savedir=None):
        """Save the CrossRegistrar object to a file."""

        if savedir is None:
            savedir = self.savedir
        self.savedir = savedir

        with open(os.path.join(savedir, 'info.pkl'), 'wb') as f:
            pkl.dump(self._todict(), f, pkl.HIGHEST_PROTOCOL)

    @classmethod
    def load(cls, path):
        return cls(None, None, path)

    def _todict(self):
        return {'savedir': os.path.abspath(self.savedir),
                'label': self.label,
                'ds_dirs': self.ds_dirs}

    def __str__(self):
        return '<CrossRegistrar>'

    def __repr__(self):
        return (f'<CrossRegistrar: num_datasets={self.num_datasets}, '
                + f'num_planes={self.num_planes}>')

    # ------------------------- Preprocessing steps ------------------------- #

    @property
    def preprocessing(self):
        try:
            return self._preprocessing
        except AttributeError:
            self._preprocessing = pd.read_hdf(
                os.path.join(self.savedir, "preprocessing.h5"))
            return self._preprocessing

    @preprocessing.setter
    def preprocessing(self, df):
        self._preprocessing = df
        if self.savedir is not None:
            df.to_hdf(os.path.join(self.savedir, "preprocessing.h5"),
                      key="data", mode="w")

    @property
    def time_averages(self):
        return self.preprocessing['time_averages']

    @property
    def crops(self):
        return self.preprocessing['crops']

    @property
    def transforms(self):
        return self.preprocessing['transforms']

    @property
    def num_rois(self):
        return self.preprocessing['num_rois']

    @property
    def reference_idx(self):
        return np.where(self.preprocessing['order'].values == 0)[0][0]

    def preprocess(self, suite2p=True, crop=True, overwrite=False):
        """Run necessary preprocessing steps to prepare datasets for
        cross-registration: identify the reference session, find and crop
        the video time averages, and compute transformations to the reference
        image coordinates for later use.

        Parameters
        ----------
        suite2p : bool, optional
            If Suite2p files exist, use the enhanced time average image from
            the ops dictionaries for each plane. Defaults to True
        crop : bool, optional
            Crop image to remove rows or columns with only constant values.
            Recommended for Suite2p processed data. Defaults to True.
        overwrite : bool, optional
            Whether to run preprocessing again if existing data already
            exists.
        """

        if not overwrite:
            try:
                self.preprocessing
                print("Object has already be preprocessed. Set overwrite=True"
                      + " to reprocess")
                return
            except Exception:
                pass

        # initialize dataframe to hold preprocessing results
        df = pd.DataFrame(index=self.ds_dirs,
                          columns=['num_rois', 'order', 'time_averages',
                                   'crops', 'transforms'])

        print("Preparing preprocessing info for registration")
        for ds_dir, df_row in df.iterrows():
            # get number of rois, time averages, and crops for each session
            ds = ImagingDataset.load(ds_dir)
            df_row['num_rois'] = len(ds.ROIs[self.label])

            if suite2p:
                try:
                    time_averages = utils.get_suite2p_time_average(ds)
                except Exception:
                    time_averages = ds.time_averages
            else:
                time_averages = ds.time_averages

            if crop:
                crops = [utils.find_crop_coords(plane)
                         for plane in time_averages]
                df_row['crops'] = crops
                time_averages = np.stack([image[rows][:, columns]
                                          for (rows, columns), image
                                          in zip(*[crops, time_averages])])
            df_row['time_averages'] = time_averages

        # determine processing order by number of ROIs per session
        df['order'] = np.argsort(np.argsort(df['num_rois'] * -1))
        reference_idx = np.where(df.order.values == 0)[0][0]
        self._reference_idx = reference_idx
        print(f"{df.iloc[reference_idx].name} has "
              + f"{df.iloc[reference_idx].num_rois} rois and will serve as "
              + "the reference session")

        # compute transforms to and from the reference session (order 0)
        transforms = []
        for plane in range(self.num_planes):
            print("Calculating transforms to reference session")
            transforms.append(utils.compute_transforms(
                [img[plane] for img in df.time_averages],
                reference_idx=reference_idx))
        df['transforms'] = list(zip(*transforms))

        self.preprocessing = df
        return self.preprocessing

    # ----------------------- Registration functions ------------------------ #

    def _get_dataset(self, order_idx):
        return ImagingDataset.load(self.preprocessing[
            self.preprocessing.order == order_idx].iloc[0].name)

    @property
    def reference_data(self):
        try:
            return self._reference_data
        except AttributeError:
            self._reference_data = pd.read_hdf(
                os.path.join(self.savedir, "reference.h5"))

    @reference_data.setter
    def reference_data(self, df):
        self._reference_data = df
        if self.savedir is not None:
            df.to_hdf(os.path.join(self.savedir, "reference.h5"),
                      key="data", mode="w")

    def _create_reference(self, force_reference=None, overwrite=False):
        """Initialize the reference dataframe

        Parameters
        ----------
        force_reference : str, optional
            Path to sima folder of reference dataset. If None (default),
            the dataset with the greatest number of cells is used as the
            reference.
        overwrite : bool, optional
            Recompile the reference from scratch
        """
        if not overwrite:
            try:
                self.reference_data
                print("Using existing reference data. Set overwrite=True to"
                      + " recompile")
                return
            except Exception:
                pass

        # find the reference dataset
        if force_reference:
            ref_ds = ImagingDataset.load(force_reference)
        else:
            ref_ds = self._get_dataset(0)
        crop = self.crops.loc[ref_ds.savedir]

        # get dataframe of ROI information and save as attribute
        reference_data = pd.DataFrame(
            [utils.get_roi_info(roi, crop=crop)
             for roi in ref_ds.ROIs[self.label]])
        reference_data['source_dataset'] = ref_ds.savedir
        reference_data['partners'] = \
            [{roi.source_dataset: label} for label, roi
             in reference_data.iterrows()]
        reference_data['mask_pairs'] = \
            [{roi.source_dataset: roi['mask']} for label, roi
             in reference_data.iterrows()]
        self.reference_data = reference_data
        return self.reference_data

    @property
    def _reference_size(self):
        return self.preprocessing.loc[
            self.reference_data.iloc[0].source_dataset].time_averages.shape

    def _pair_dataset(self, idx, pairing_threshold=0.5, max_distance=25):
        """Do pairing for dataset against the current reference stack

        Parameters
        ----------
        pairing_threshold : float, optional
        max_distance : int, optional
        """

        ds = self._get_dataset(idx)
        if ds.savedir in self.reference_data.source_dataset.values:
            print(f"'{ds.savedir}' is already paired, or was the reference "
                  + "session. Skipping...")
            return

        print(f"Pairing {self.num_rois[ds.savedir]} ROIs from {ds.savedir}")

        print(" -- Transforming ROIs to reference")
        rois = pd.DataFrame(
            [utils.get_roi_info(roi, crop=self.crops[ds.savedir],
                                transform=self.transforms[ds.savedir],
                                target_shape=self._reference_size)
             for roi in ds.ROIs[self.label]])

        reference_data = self.reference_data

        for plane in np.unique(rois.plane):
            print(f" -- Plane {plane}")

            plane_rois = rois[rois.plane == plane]
            reference_rois = reference_data[reference_data.plane == plane]
            print(f" ---- Pairing {len(plane_rois)} ROIs to"
                  + f" {len(reference_rois)} references")

            print(" ---- Computing distance matrix")
            dist_mat = utils.calc_distance_matrix(reference_rois['mask'],
                                                  plane_rois['mask'])

            # binarize and find optimal pairs via Hungarian algorithm
            print(" ---- Finding optimal pairs")
            dist_mat[dist_mat > pairing_threshold] = 1  # binarize at threshold
            roi_pairs = linear_sum_assignment(dist_mat)
            roi_pairs = [(reference_rois.iloc[x].name, plane_rois.iloc[y].name)
                         for x, y in zip(*roi_pairs)
                         if dist_mat[x, y] < pairing_threshold]

            print(f" ---- Autopaired {len(roi_pairs)}/{len(plane_rois)} ROIs")
            for ref_label, new_label in roi_pairs:
                reference_data.loc[ref_label].partners[ds.savedir] = new_label
                plane_rois.drop(index=new_label, inplace=True)

            print(f" ---- Starting manual review of remaining ROIs")
            total_unpaired = len(plane_rois)
            for uidx, (label, unpaired_roi) in \
                    enumerate(plane_rois.iterrows()):
                print(f"Reviewing ROI {uidx + 1}/{total_unpaired}")

                eligible_ref_rois = utils.find_nearby_rois(
                    unpaired_roi.centroid, reference_rois,
                    max_distance=max_distance, exclude_partnered=ds.savedir)

                if eligible_ref_rois.empty:
                    print(f"No nearby ROIs, adding as new entry to reference")
                else:
                    ref_label = self._plot_manual_pair(
                        unpaired_roi, eligible_ref_rois,
                        max_distance=max_distance)

                    if ref_label is not None:
                        if ref_label == -1:
                            print("Rejected ROI")
                        else:
                            reference_data.loc[ref_label].partners[
                                ds.savedir] = unpaired_roi.name
                            reference_data.loc[ref_label].mask_pairs[
                                ds.savedir] = unpaired_roi['mask']
                        plane_rois.drop(index=unpaired_roi.name, inplace=True)
                        continue
                    else:
                        print("No matching ROIs, adding as new entry to "
                              + "reference")

                # add unpaired ROI as a new entry to reference if a pair was
                # not identified
                unpaired_roi['source_dataset'] = ds.savedir
                unpaired_roi['partners'] = {ds.savedir: unpaired_roi.name}
                unpaired_roi['mask_pairs'] = {ds.savedir: unpaired_roi['mask']}
                reference_data.loc[unpaired_roi.name] = unpaired_roi
                plane_rois.drop(index=unpaired_roi.name, inplace=True)

        # resave reference
        print("ROIs registered for all planes.\n"
              + f"Total count is {len(reference_data)}.\n"
              + "ROIs appear in {} ".format(
                np.mean([len(p) for p in reference_data.partners]))
              + "sessions on average.\n"
              + f"Updating reference for future registrations")
        self.reference_data = reference_data
        return self.reference_data

    def _is_in_bounds(self, roi):
        mask = roi['mask']
        for source_dataset, session in self.preprocessing.iterrows():
            if source_dataset == roi.source_dataset:
                continue

            new_mask = utils.apply_transform(
                mask, session.transforms[roi.plane],
                session.time_averages.shape[1:], invert=True)

            if new_mask.sum() == 0:
                return False
        return True

    def remove_out_of_bound_rois(self):
        reference_data = self.reference_data
        print("Checking list of {} ROIs for out-of-bounds masks".format(
            len(reference_data)))
        for label, roi_info in reference_data.iterrows():
            if not self._is_in_bounds(roi_info):
                print(f"...removed {label}")
                reference_data.drop(index=label, inplace=True)
        self.reference_data = reference_data
        print(f"{len(reference_data)} ROIs remain")
        return self.reference_data

    def register_rois(self, pairing_threshold=0.5, max_distance=25,
                      remove_out_of_bound_rois=True,
                      force_reference=None, overwrite=False,
                      preprocessing_kws={}):
        """Main registration function. Iterate over ImagingDatasets and
        pair ROIs with the reference dataset. As new ROIs are found, they are
        added to the reference. The augmented reference set is then used to
        pair the next dataset, and so the reference grows over time and
        facilitates additional autopairing.

        Parameters
        ----------
        pairing_threshold : float [0, 1], optional
            Jaccard score threshold for autopairing. ROI pairs with scores
            below (?) this threshold will be flagged for manual review.
            Defaults to 0.5
        max_distance : float, optional
            Maximum x and y distance in pixels between ROI centroids for
            considering candidate reference ROIs to pair under manual review
        remove_out_of_bound_rois : bool, optional
            After pairing, remove any ROIs that would fall outside of the FOV
            in any single session. We apply the inverse transform (from
            the reference to each session) on each ROI in the final list and
            make sure the masks are non-empty in every session). Defaults to
            True.
        force_reference : str, optional
            ImagingDataset.savedir of dataset to force as the reference.
            Defaults to None, in which case the dataset with the greatest
            number of ROIs is chosen.
        overwrite : bool, optional
            If False, registration will resume from previous work. If True,
            prior results will be cleared and registration will start from
            scratch.
        preprocessing_kws : dict
            Pass additional arguments to the preprocessing pipeline.
        """

        self.preprocess(overwrite=overwrite, **preprocessing_kws)
        self._create_reference(force_reference=force_reference,
                               overwrite=overwrite)

        for idx in range(len(self.ds_dirs)):
            self._pair_dataset(idx, pairing_threshold=pairing_threshold,
                               max_distance=max_distance)

        if remove_out_of_bound_rois:
            self.remove_out_of_bound_rois()

    # ------------------------- Plotting functions -------------------------- #

    def _plot_manual_pair(self, unpaired_roi, eligible_ref_rois,
                          max_distance=25):
        """Generates an interactive plot to choose the correct ROI"""

        time_avg = self.time_averages[
            eligible_ref_rois.iloc[0].source_dataset][unpaired_roi.plane]
        ylims, xlims = [[-max_distance, max_distance] + c
                        for c in unpaired_roi.centroid]

        plt.figure(figsize=(6, 6))
        sns.heatmap(time_avg, square=True, cbar=False, cmap='Greys_r')

        mask_stack = np.stack(eligible_ref_rois['mask']).sum(axis=0)
        sns.heatmap(mask_stack, mask=(mask_stack == 0), alpha=0.4, cmap='Reds',
                    cbar=False)
        for num, centroid in enumerate(eligible_ref_rois.centroid):
            text = plt.text(centroid[1], centroid[0], num, fontsize=25,
                            horizontalalignment='center',
                            verticalalignment='center', color='red',
                            weight='bold')
            text.set_path_effects([path_effects.Stroke(linewidth=1,
                                                       foreground='black'),
                                   path_effects.Normal()])

        sns.heatmap(unpaired_roi['mask'],
                    mask=~unpaired_roi['mask'].astype('bool'),
                    vmin=0, vmax=1, alpha=0.4, cmap='Blues', cbar=False)

        plt.title('Available ROIs')
        plt.xlim(xlims)
        plt.ylim(ylims)
        plt.axis('off')
        plt.show()

        # let the user pick which cell to retain
        while True:
            decision = input(
                "Choose the matching ROI\n(Press enter if new, -1 to discard)")
            try:
                if decision == '':
                    clear_output()
                    return None
                elif (int(decision) in np.arange(len(eligible_ref_rois))):
                    clear_output()
                    return eligible_ref_rois.iloc[int(decision)].name
                elif int(decision) == -1:
                    clear_output()
                    return -1
                else:
                    print('Invalid input!')
            except ValueError:
                print("Invalid input!")

    def _plot_avg_mask(self, roi):
        # TODO
        pass

    def _plot_cumulative_fov(self):
        # TODO
        pass

    # ---------------------- Signal splitting functions --------------------- #

    def update_signals(self, overwrite=False):
        """Update signals dataframes for the underlying datasets. A new
        ROI list is stored that contains all ROIs (whether they appear in
        a given session or not). If an ROI is not present in a particular
        session, its trae appears in the signal dataframes as NaN.

        Parameters
        ----------
        overwrite : bool, optional
            Whether to overwrite the registered ROIs, if they already exist.
            Defaults to True.
        """
        for sima_dir in self.preprocessing.index:
            print(f"Saving results for {sima_dir}")

            ds = ImagingDataset.load(sima_dir)
            original_rois = ds.ROIs[self.label]
            if len(original_rois) == 0:
                print(" - Session has no ROIs, skipping...")
                continue

            with SignalFile(os.path.join(sima_dir, "signals.h5")) \
                    as signal_file:

                # find all signals dataframes for self.label
                keys_to_update = [key for key in signal_file.keys()
                                  if f'/{self.label}/' in key]
                for key in keys_to_update:
                    new_key = key.replace(f'/{self.label}/',
                                          f'/{self.label}_registered/')
                    print(f" - Updating {key} to {new_key}")
                    if new_key in signal_file:
                        if not overwrite:
                            print(f"Signals already exist for {new_key}. Set "
                                  + "overwrite=True to re-import. Skipping...")
                            continue
                        print(f" - Deleting previous import at {new_key}")
                        signal_file.remove(new_key)

                    # compile new dataframe with all ROIs
                    df = signal_file.get(key)
                    new_df = pd.DataFrame(index=[], columns=df.columns)
                    new_df.index.name = df.index.name
                    for new_label, roi_info in self.reference_data.iterrows():
                        try:
                            # find the name of the ROI in the current session
                            new_df.loc[new_label] = \
                                df.loc[roi_info.partners[sima_dir]]
                        except KeyError:
                            new_df.loc[new_label] = np.nan

                    # save
                    signal_file.put(new_key, new_df)

            # add ROI list to the dataset
            # TODO - should we correct the masks to be of the right dimensions
            #        for the given dataset?
            print(" - Adding ROIList to dataset")
            ds.add_ROIs(ROIList(self.reference_data.roi.values),
                        f'{self.label}_registered')

    # -------------------------- Cleanup functions -------------------------- #



import os
import argparse
import numpy as np
import warnings
import time
import sys
import multiprocessing
import itertools as it
import warnings
import copy
import shutil

from sima import Sequence
from sima import ImagingDataset

from progressbar import ProgressBar
from sima.sequence import _IndexedSequence
from sima.sequence import _MotionCorrectedSequence
#from sima.sequence import _TimeAvgSequence
from sima.motion import PlaneTranslation2D


def _first_obs(seq):
    print('searching for first observed values for each pixel')
    p = ProgressBar(seq.shape[0])
    p.start()

    frame_iter1 = iter(seq)
    first_obs = next(frame_iter1)
    for i, frame in enumerate(frame_iter1):
        p.update(i)
        for frame_chan, fobs_chan in zip(frame, first_obs):
            fobs_chan[np.isnan(fobs_chan)] = frame_chan[np.isnan(fobs_chan)]
        if all(np.all(np.isfinite(chan)) for chan in first_obs):
            break
    p.finish()
    return first_obs


def _fill_gaps(first_obs, frame_iter2):
    """Fill missing rows in the corrected images with data from nearby times.

    Parameters
    ----------
    frame_iter1 : iterator of list of array
        The corrected frames (one list entry per channel).
    frame_iter2 : iterator of list of array
        The corrected frames (one list entry per channel).

    Yields
    ------
    list of array
        The corrected and filled frames.

    """
    most_recent = [x * np.nan for x in first_obs]
    for i, frame in enumerate(frame_iter2):
        for fr_chan, mr_chan in zip(frame, most_recent):
            mr_chan[np.isfinite(fr_chan)] = fr_chan[np.isfinite(fr_chan)]
        yield np.array([np.nan_to_num(mr_ch) + np.isnan(mr_ch) * fo_ch
               for mr_ch, fo_ch in zip(most_recent, first_obs)])


def frame_blocks(frames, chunk_size):
    try:
        t1, t2 = frames
    except TypeError:
        t1 = 0
        t2 = frames

    n_frames = t2-t1
    blocks = np.linspace(
        t1, t2, int(np.ceil(float(n_frames)/chunk_size))+1,
        dtype=int)
    return list(zip(blocks[:-1], blocks[1:]))


def pool_helper(args):
    return args[0](*args[1:])


def frame_writer(out_file, first_obs, dtype, seq_dictionary):
    seq_dictionary = copy.deepcopy(seq_dictionary)
    seq = seq_dictionary.pop('__class__')._from_dict(seq_dictionary)
    num_f = seq._base.shape[0]

    #iterator = _fill_gaps(first_obs, iter(seq))
    iterator = iter(seq)
    shape = (np.prod(seq.shape[1:]), seq.shape[0])
    start = seq._indices[0].start

    f = np.memmap(out_file, mode='r+', dtype=dtype, order='C', # or C? who knows...
                  shape=(shape[0], num_f))

    p = ProgressBar(seq.shape[0])
    p.start()
    Z = 0
    Y = 1
    X = 2 
    C = 3

    for f_idx, frame in enumerate(iterator):
        p.update(f_idx)
        f[:,start+f_idx] = np.clip(
            np.nan_to_num(np.reshape(seq._get_frame(f_idx).transpose((Z, X, Y, C)), 
                shape[0], order='C')), 0, None) # .transpose((C,X,Y,Z))
    p.finish()
    del f
    return out_file


def convert(sequence, out_file, cycle=0, chunk_size=500, n_processes=4, channel='Ch2', dtype=np.single,
            first_obs=150, trim=0, z_plane=None, skip_init=0, n_frames=None, order='C', ):
            # time_avg=False?
    #ds = ImagingDataset.load(in_file)
    #channel = ds._resolve_channel(channel)
    # if time_avg:
    #     sequence = _TimeAvgSequence(ds.sequences[cycle])[:n_frames]
    # else:

        # TODO: multiple sequences
        # TODO: does this handle z correctly?
    #sequence = ds.sequences[0][:n_frames]
    shape = (np.prod(sequence.shape[2:]), sequence.shape[0])
    #seq_dictionary = sequence._todict()
    first_obs = _first_obs(sequence[:first_obs])

    blocks = frame_blocks(shape[1], shape[1]/n_processes)
    sequences = []
    for t1,t2 in blocks:
        sequences.append(sequence[t1:t2]._todict())
#    del ds
    del sequence

    f = np.memmap(out_file, mode='w+', dtype=dtype, order=order, shape=shape)

    # Need this line to avoid acquiring a lock that cannot be released (deadlock) -Z
    with multiprocessing.get_context("spawn").Pool(processes=n_processes) as pool:
        start = time.time()
        num_blocks = len(blocks)
        print('exporting in %i blocks' % (num_blocks))
        pool.map(
            pool_helper,
            zip(it.repeat(frame_writer), it.repeat(out_file), it.repeat(first_obs), it.repeat(dtype),
                sequences))
    print('finished in %i seconds' % (time.time() - start))


def main():
    argParser = argparse.ArgumentParser()
    argParser.add_argument(
        "-s", "--scratch_drive", action="store", type=str,
        default="/scratch/caiman", help="destination drive for memmap file")
    argParser.add_argument(
        "-c", "--channel", action="store", type=str,
        default="Ch2", help="channel name. default is Ch2")
    argParser.add_argument(
        '-n', '--num_frames', action='store', type=int,
        default=None, help='number of frames to convert')
    argParser.add_argument(
        '-t', '--time_avg', action='store_true', default=False,
        help='apply _TimeAvgSqeuence wrapper prior to conversion')
    argParser.add_argument(
        "filename", action="store", type=str, default="",
        help=("sima file to convert to memmap"))
    args = argParser.parse_args()

    ext, filename = os.path.split(args.filename)
    scratch_dir = os.path.join(args.scratch_drive, ext[1:])
    if not os.path.exists(scratch_dir):
        os.makedirs(scratch_dir)

    savefile = os.path.join(
        scratch_dir, f'{os.path.splitext(filename)[0]}.mmap')

    convert(args.filename, savefile,
            channel=args.channel, first_obs=150, n_frames=args.num_frames)

    #sequence = ImagingDataset.load(args.filename).sequences[0]
    ds = ImagingDataset.load(args.filename)
    if args.time_avg:
        sequence = _TimeAvgSequence(ds.sequences[0])[:args.num_frames]
    else:
        sequence = ds.sequences[0][:args.num_frames]

    print(sequence.shape)
    seq = Sequence.create(
        'memmap', savefile,(sequence.shape[2], sequence.shape[3], sequence.shape[0], 1),
        'yxtc')
    sima_file = os.path.join(scratch_dir, filename)
    if os.path.exists(sima_file):
        shutil.rmtree(sima_file)

    print(f'memmap file:\n{savefile}')
    print(f'sima filepath:\n{sima_file}')
    ImagingDataset([seq], sima_file)


if __name__ == '__main__':
    main()



import numpy as np
import pandas as pd

from lab3.core import Analysis
from lab3.filters import IndexIdentity, ColumnIdentity

class Histogram(Analysis):
    constructor_ids = ['lower', 'upper', 'nbins', 'bins']
    
    def __init__(self, lower=0., upper=1., nbins=100, bins=None):
        if bins is None:
            self.lower = lower
            self.upper = upper
            self.nbins = nbins
            self.bins = np.linspace(lower, upper, nbins+1)
        else:
            self.lower = bins[0]
            self.upper = bins[-1]
            self.nbins = len(bins)
            self.bins = bins
            
        self.bin_size = self.bins[1] - self.bins[0]
    
    def to_columns(self):
        return pd.Index(self.bins[:-1], name=self.name)
    
    def apply_to(self, expt, signal_spec, roi_filter=IndexIdentity, time_filter=ColumnIdentity):
        signals = expt.signals(**signal_spec)
        signals = signals.loc[roi_filter.bind_to(expt)]
        signals.loc[:, ~time_filter.bind_to(expt)] = np.nan
        return self.calculate(signals)
        
    def calculate(self, signals):
        clean_signals = signals.dropna(axis=1)
                
        result = clean_signals.apply(lambda x: np.histogram(x, bins=self.bins)[0], axis=1)
        return pd.DataFrame((item for item in result), index=result.index, columns=self.columns)



import os
import shutil
import warnings

from abc import abstractmethod, ABCMeta
from sima import ImagingDataset
from pkg_resources import get_distribution

import matplotlib.pyplot as plt
import numpy as np
import psutil
from scipy.ndimage.filters import gaussian_filter
import sys
import pickle as pkl

import caiman as cm
import caiman.source_extraction.cnmf as cnmf
from caiman.components_evaluation import evaluate_components, estimate_components_quality_auto
from caiman.cluster import setup_cluster
from caiman.paths import caiman_datadir

from lab3.extraction.cnmf import convertToMemmap, AROIs, extract_signals, import_signals
from lab3.misc.sima_compatibility import sima_compatible
from lab3.misc.sima_sequences import _NoRMCorreRigidSequence

FASTSCRATCH = "/fastscratch/memmaps"

@sima_compatible
class CNMFImagingDataset(ImagingDataset):
    def extract(self, overwrite=False, reprocess=True, on_scratch=True, 
        first_obs=0, signal_channel='Ch2', n_processes=8, 
        # NormCorre arguments
        use_normcorre=True, max_shifts=(6, 6), strides=(48, 48), overlaps=(24, 24), 
        pw_second_pass=False, normcorre_kws={},
        # CNMF arguments
        cnmf_second_pass=True, rf=15, stride=10, K=12, gSig=(2, 2), gnb=2,
        merge_thresh=0.8, p=0, fr=10, decay_time=1., cnmf_kws={}):

        # TODO: document
        self.write_memmap(overwrite=overwrite, reprocess=reprocess, 
                            on_scratch=on_scratch, first_obs=first_obs, 
                            signal_channel=signal_channel, n_processes=n_processes)

        c, dview, n_processes = setup_cluster(
            backend='local', n_processes=n_processes, single_thread=False)

        try:
            if use_normcorre:
                self.motion_correct(dview, max_shifts=max_shifts, strides=strides, 
                        overlaps=overlaps, pw_second_pass=pw_second_pass, **normcorre_kws)

            print("Extracting with CNMF...")
            self.cnm = extract_signals.extract(self.mmapname, dview, n_processes=n_processes, 
                                                cnmf_second_pass=cnmf_second_pass, rf=rf, 
                                                stride=stride, K=K, gSig=gSig, gnb=gnb,
                                                merge_thresh=merge_thresh, 
                                                p=p, fr=fr, decay_time=decay_time, 
                                                **cnmf_kws)
            self.save_cnmf_results()
        except:
            raise
        finally:
            # Whatever happens, close the cluster
            cm.stop_server(dview=dview)

    def extract3D(self, max_shifts=(8, 8, 2), strides=(48, 48, 4), overlaps=(24, 24, 2), 
                    gSig=(2, 2, 2.5), **kwargs):
        """Identical to `extract`, just with default parameters set for 3D
        """

        self.extract(max_shifts=max_shifts, strides=strides, overlaps=overlaps, 
                    gSig=gSig, **kwargs)

    def write_memmap(self, overwrite=False, reprocess=True, on_scratch=True, 
                    first_obs=50, signal_channel='Ch2', n_processes=8):
        if on_scratch:
            self.mmapname = os.path.join(FASTSCRATCH, self.mmapname)

        # TODO: convert assumes sequence 0
        if not os.path.exists(self.mmapname) or overwrite:
            convertToMemmap.convert(self.savedir, self.mmapname, n_processes=n_processes,
                channel=signal_channel, first_obs=first_obs, n_frames=self.num_frames)
        elif reprocess:
            pass # use existing mmap
        else:
            raise OSError(f"`{self.mmapname}` already exists (reprocess=True to extract from "\
                            f"existing .mmap or overwrite=True to overwrite)")

    def motion_correct(self, dview, max_shifts=(6, 6), strides=(48, 48), 
                        overlaps=(24, 24), pw_second_pass=False, **normcorre_kws):

        is3D = self.frame_shape[0] > 1

        if is3D:
            assert len(max_shifts) == len(strides) == len(overlaps) == 3
            print("3D check passed")
        else:
            assert len(max_shifts) == len(strides) == len(overlaps) == 2

        print("Motion correcting...")
        self.mc = extract_signals.motion_correct(self.mmapname, dview=dview, 
                    max_shifts=max_shifts, strides=strides, overlaps=overlaps, 
                    pw_second_pass=pw_second_pass, is3D=is3D, **normcorre_kws)
        self.mmapname, = self.mc.mmap_file
        print("Done.")
        print(f"MC file: {self.mmapname}")

    def import_results(self, **kwargs):
        """Import Suite2p signal results using the new-style signal formats
        (dataframes and h5 stores).

        Parameters
        ----------
        ds : CNMFImagingDataset
            Instance of a Suite2pStrategy, which has been previously extracted.
        label : str, optional
            ROI label to store the signals and ROI masks under. Defaults to
            'cnmf'.
        channel : {str or int}, optional
            Name or index of dynamic channel in the imaging dataset. Defaults
            to 'Ch2'.
        overwrite : bool, optional
            If key '/{channel}/{label}' already exists in the dataset signals
            file, overwrite with new import. Note this will delete any
            additional signal types in the store in addition to the 'raw' and
            'npil' (e.g. 'dfof'), to prevent mixtures of signals derived from
            different imports. Defaults to False, which will raise an error if
            the desired key already exists.

        See also
        --------
        lab3.extraction.import_signals.import_to_signals_file
        """

        import_signals.import_to_signals_file(self, **kwargs)

    def apply_mc_results(self, write=False):
        """Apply displacements calculated by NormCorre to the underlying dataset.
        Currently only supports rigid motion correction.

        Parameters
        ----------
        self : CNMFDataset
            Instance of a CNMFDataset, which has been previously extracted.
        overwrite : bool, optional
            Whether to save the motion corrected dataset as a sima directory.
        """

        new_sequences = []

        for seq in self.sequences:
            mc_seq = _NoRMCorreRigidSequence(seq, self.mc)
            new_sequences.append(mc_seq)

        self.sequences = new_sequences

        if write:
            # backup_dir(self.savedir, delete_original=False)
            self.save()

    def save_cnmf_results(self):
        cnmf_pkl = os.path.join(self.savedir, 'cnm.pkl')

        # Make pickleable
        self.cnm.dview = None 
        self.mc.dview = None 

        with open(cnmf_pkl, "wb") as cnmf_file:
            print("Saving cnmf results...")
            pkl.dump(self.cnm, cnmf_file)

        if hasattr(self, 'mc'):
            print("Saving mc results...")
            mc_pkl = os.path.join(self.savedir, 'normcorre_mc.pkl')
            with open(mc_pkl, "wb") as normcorre_file:
                pkl.dump(self.mc, normcorre_file)

    @property 
    def dims(self):
        return self.cnm.dims

    @property
    def cnm(self):
        try:
            return self._cnm
        except AttributeError:
            cnmf_pkl = os.path.join(self.savedir, 'cnm.pkl')
            with open(cnmf_pkl, "rb") as cnmf_file:
                self._cnm = pkl.load(cnmf_file)
            return self._cnm 

    @cnm.setter 
    def cnm(self, value):
        self._cnm = value

    @property
    def mc(self):
        try:
            return self._mc
        except AttributeError:
            mc_pkl = os.path.join(self.savedir, 'normcorre_mc.pkl')
            with open(mc_pkl, "rb") as normcorre_file:
                self._mc = pkl.load(normcorre_file)
            return self._mc
    
    @mc.setter 
    def mc(self, value):
        self._mc = value

    @property
    def mmapname(self):
        try:
            return self._mmapname
        except AttributeError:
            z, y, x, c = self.frame_shape
            t = self.num_frames
            mmapname = f"Yr_d1_{y}_d2_{x}_d3_{z}_order_C_frames_{t}_.mmap"
            self._mmapname = mmapname
            return mmapname

    @mmapname.setter 
    def mmapname(self, value):
        self._mmapname = value


"""The extraction module contains structures for handling signal extraction
from imaging datasets, and related problems (identifying ROIs,
calculating traces, etc...).

Currently implemented are routines for using the Suite2p package
(lab3.extraction.s2p). In the future, this could be expanded to handle other
pipelines (e.g. CaImAn, manual segmentation...)"""



import pandas as pd
import numpy as np
from functools import reduce
import warnings

from .classes import Item, Group
from .helpers import wrap

class Metadata(Item):
    """Base class for adding metadata to a dataframe, most likely a signals
    dataframe. 

    Example usecases include: annotating a dendritic imaging dataset with morphological information;
    adding ROI identities determined by post hoc immunohistochemistry to functional datasets; 
    marking cells identified on the red channel; annotating cells as place and non-place cells

    Metadata objects can be layered using the |, &, or + operators (same behavior).

    Examples
    --------
    Add metadata to a signals dataframe:

    >>> meta = SomeMetadata(...)
    >>> meta2 = OtherMetadata(...)
    >>> annotated_signals = expt.signals(signal_type='dfof', label='PYR', metadata = meta | meta2)

    Perform analysis on an annotated dataframe

    >>> signal_spec = {'signal_type': 'dfof', 'label':'PYR', 
    >>>                 'metadata': immuno_data | is_place_cell | is_active}
    >>> result = my_cohort.apply(my_analysis, signal_spec=signal_spec)
    """

    levels = []

    def to_levels(self):
        return self.levels

    def bind(self, expt):
        """Retrieve the metadata from the Experiment object

        Returns
        -------
        pd.Index or pd.MultiIndex
        """
        raise NotImplementedError
    
    def bind_to(self, signals, expt):
        """Retrieve the metadata from the Experiment object, and add it to the signals
        DataFrame.

        Note: This implementation assumes that the output of `bind` is in one-to-one 
        correspondence with the signals DataFrame index. If more sophisticated 
        alignment is needed, this method needs to be overriden. 

        See lab3.metadata.base for examples
        """
        try:
            index = self.bind(expt)
            try:
                return signals.set_index(index, append=True)
            except ValueError:
                # Wrap strategy
                assert len(index) == 1
                dataframe = signals.copy()

                if len(index.names) == 1:
                    return wrap(dataframe, index[0], index.name)
                else:
                    for key, level in zip(index[0], index.names):
                        dataframe = wrap(dataframe, key, header=level)
                    return dataframe

        except Exception as e:
            warnings.warn(f"Unable to add metadata `{str(self)}` because of {e}")
            return signals
    
    def __and__(self, other):
        return MetadataGroup([self, other])
    
    def __add__(self, other):
        return self & other
    
    def __or__(self, other):
        return self & other
    
    def __str__(self):
        return self.name
    
    def _repr_html_(self):
        return pd.DataFrame([], index=pd.MultiIndex.from_frame(
            pd.DataFrame([], columns=self.to_levels())))._repr_html_()

class MetadataGroup(Metadata, Group):
    def to_levels(self):
        return sum([m.to_levels() for m in self], [])

    def bind(self, expt):
        return _deduplicate(pd.MultiIndex.from_frame(
            _join_indexes([m.bind(expt) for m in self])))
    
    def bind_to(self, signals, expt):        
        return _deduplicate_index(reduce(lambda x, y: y.bind_to(
            x.bind_to(signals, expt), expt), self))
    
    def __str__(self):
        return " | ".join([str(m) for m in self])
    
class _DefaultMetadata(Metadata):
    def bind(self, expt):
        return pd.Index([])
    
    def bind_to(self, signals, expt):
        return signals

    def __call__(self, *args, **kwargs):
        return self

DefaultMetadata = _DefaultMetadata()

def _join_indexes(indexes):
    return reduce(lambda x,y: x.join(y.to_frame(index=False)), indexes, 
           pd.DataFrame([[]]))

def _deduplicate(index):
    index_names = pd.Index(index.names)
    where_duplicated = np.where(index_names.duplicated())[0]
    return index.droplevel(level=list(where_duplicated))

def _deduplicate_index(dataframe):
    df = dataframe.copy()
    index = dataframe.index
    dedup = _deduplicate(index)
    df.index = dedup
    return df



from abc import ABCMeta, abstractmethod
from itertools import repeat
from multiprocessing import Pool

import numpy as np
import pandas as pd

from lab3.analysis.misc import find_intervals
from .spatial_tuning import SpatialTuningStrategy, CircularMixin
from ..base import BaseSignalAnalysis


class PlaceFieldMeta(ABCMeta):

    """Custom metaclass for place field strategies, to force the user to
    implement certain attributes."""

    abstract_attributes = []

    def __call__(self, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)
        for attr_name in obj.abstract_attributes:
            if not getattr(obj, attr_name):
                raise ValueError(f'Required attribute ({attr_name}) not set')
            if attr_name == 'tuning_strategy':
                ts = getattr(obj, attr_name)
                assert isinstance(ts, SpatialTuningStrategy), \
                    f"{ts} is not a valid SpatialTuningStrategy"
        return obj


class PlaceFieldStrategy(BaseSignalAnalysis, metaclass=PlaceFieldMeta):

    """Abstract base class for implementing place field detection algorithms.
    """

    abstract_attributes = ['tuning_strategy'] #, 'metric_functions']

    @property
    def circular(self):
        """Returns True if self.tuning_strategy implies a circular track.
        In this case we allow place fields to wrap around the ends of the
        track."""
        return isinstance(self.tuning_strategy, CircularMixin)

    def detect(self, signals, position, laps=None, sample_filter=None):
        """This is the main method for place field detection.

        Parameters
        ----------
        signals : pd.DataFrame
            Typically an array shape (n_rois, n_samples), keyed by the ROI
            label. Exact form of input depends on the tuning strategy used.
        position : array (n_samples)
            Array containing the integer position bin for every sample in
            signals
        laps : array (n_samples)
            Array containing the lap id of ever sample in signals.
        sample_filter : array (n_samples)
            Boolean array. Sample indices that are False are ignored during
            detection.

        Returns
        -------
        place_fields : pd.DataFrame (variable length)
            This should be a multi-index dataframe with top index roi_label
            and second index counting each place field. The columns should
            include...?
        """

        if sample_filter is not None:
            signals = signals.loc[:, sample_filter]
            position = position[sample_filter]
            laps = laps[sample_filter]

        return self._detect(signals, position, laps)

    def apply_metrics(self):
        try:
            return self._calculate_metrics(self)
        except AttributeError:
            raise Exception("You must detect place fields before applying "
                            + "metrics")

    # ----------------- Implement these methods to subclass ----------------- #

    @abstractmethod
    def _detect(self):
        """doc string"""
        pass

    @abstractmethod
    def _calculate_metrics(self):
        """doc string"""
        pass


def _shuffle(inputs):
    """Multiprocessing helper function"""
    np.random.seed()
    strategy_instance, signals, position, lap = inputs
    return strategy_instance.calculate(signals, position, lap, shuffle=True)


class SimplePFDetector(PlaceFieldStrategy):

    """Detect place fields by comparing the empirical tuning curve to a null
    distribution of tuning curves. Candidate place fields are spatial intervals
    where a neuron's firing rate exceeds the null distribution threshold,
    determined for the desired p-value. Place fields are then refined based
    on some simple secondary criteria: minimum and maximum allowable place
    field sizes, and a minimum number of trials that the cell fires within
    the candidate field.

    The manner in which the null distribution is sampled depends on the choice
    of SpatialTuningStrategy, which provides an additional level of
    customization to the detection beyond tuning the secondary parameters. As
    a result, this class is highly flexible and many detection schemes are
    possible.

    Parameters
    ----------
    tuning_strategy : SpatialTuningStrategy
        Instance of a SpatialTuningStrategy to be used for calculating
        tuning curves
    num_shuffles : int, optional
        How many shuffle iterations to run for generating shuffled tuning
        curves.
    pval_threshold : float, optional
        Threshold for identifying putative place fields, by comparing the
        spatial tuning curve at each bin to the null distribution. Defaults
        to 0.05 (i.e., firing rate must exceed 95% of null tuning curves at
        that bin)
    min_field_width, max_field_width : float, optional
        Minimum and maximum allowable widths to accept place fields, as a
        fraction of the track. Defaults to 0 and 1 respectively (no limits)
    min_trials_active : int, optional
        Only retain place fields wherein the cell fired on at least this many
        laps. Defaults to None (all fields are retained). Requires a
        `tuning_strategy` that implements `calculate_rasters`.
    n_processes : int, optional
        Number of workers to spawn in the parallel pool. Defaults to no
        parallelization

    Attributes
    ----------
    tuning_curves : pd.DataFrame (num_rois, num_position_bins)
    null_curves : pd.DataFrame (n_rois * num_shuffles, num_position_bins)
        Multi-indexable as null_curves.loc[(roi_label, shuffle_count)]
    threshold_curves : pd.DataFrame (num_rois, num_position_bins)
        The threshold activity rate at each bin to reach the desired
        significance level
    place_fields : pd.DataFrame (variable length)
        Multi-indexable as place_fields.loc[(roi_label, place_field_count)]
        Not all ROIs will have place fields, and some ROIs may have multiple
        fields

    Examples
    --------
    >>> # generate some fake data
    >>> signals = pd.DataFrame(np.random.normal(0, 1, (10, 1000)))
    >>> laps = np.concatenate([np.ones(100) * lap for lap in range(10)])
    >>> position = np.concatenate([np.arange(100)] * 10)
    >>>
    >>> # choose a tuning strategy
    >>> from lab3.analysis.spatial import SimpleSpatialTuning
    >>> tuning_strategy = SimpleSpatialTuning(sigma=3)
    >>>
    >>> # create detector object and detect place fields
    >>> detector = SimplePFDetector(tuning_strategy, pval_threshold=0.01,
    >>>                             min_field_width=0.05, max_field_width=0.5,
    >>>                             min_trials_active=5, num_shuffles=1000,
    >>>                             n_processes=8)
    >>> place_fields = detector.detect(signals, position, laps=laps)

    If the track is circular, we just need to pass a circular tuning strategy,
    and the detector will account for the boundary conditions.

    >>> from lab3.analysis.spatial import SimpleSpatialTuning
    >>> tuning_strategy = CircularSimpleSpatialTuning(sigma=3)
    >>> place_fields = SimplePFDetector(tuning_strategy).detect(
    >>>     signals, position, laps=laps)
    """


    def __init__(self, tuning_strategy, num_shuffles=10, pval_threshold=0.05,
                 min_field_width=0, max_field_width=1, min_trials_active=None,
                 n_processes=1):
        self.tuning_strategy = tuning_strategy
        self.num_shuffles = num_shuffles
        self.pval_threshold = pval_threshold
        self.min_field_width = min_field_width
        self.max_field_width = max_field_width
        self.min_trials_active = min_trials_active
        self.n_processes = n_processes

    def _get_nulls(self, signals, position, laps):
        """Use the inputs and self.tuning_strategy to sample from the null
        distribution of shuffle tuning curves"""

        # sample from null distribution
        iterator = repeat((self.tuning_strategy, signals, position, laps),
                          self.num_shuffles)
        if self.n_processes > 1:
            p = Pool(self.n_processes)
            null_curves = p.map(_shuffle, iterator)
            p.close()
            p.join()
        else:
            null_curves = map(_shuffle, iterator)

        # store nulls in multi-index dataframe
        null_curves = pd.concat(
                {n_idx: n for n_idx, n in enumerate(null_curves)},
                names=['shuffle', 'roi_label']
            ).swaplevel(0, 1).sort_index()

        # store as attributes
        self.null_curves = null_curves

        return null_curves

    def _get_thresholds(self):
        """Determine the threshold activity level at each spatial bin to
        establish significant spatial modulation at that location at
        p < self.pval_threshold"""

        # calculate thresholds
        thresholds = self.null_curves.groupby('roi_label').apply(
            np.percentile, axis=0, q=100 * (1 - self.pval_threshold))
        thresholds = pd.DataFrame.from_dict(
            dict(zip(thresholds.index, thresholds.values))).T
        thresholds.index.name = 'roi_label'
        thresholds.columns.name = 'position'

        # store as attributes
        self.threshold_curves = thresholds

        return thresholds

    def _find_fields(self, tuning_curves=None, threshold_curves=None):
        """Using self.tuning_curves and self.thresholds, evaluate for each
        neuron the spatial intervals along the tuning curve that exceed
        the significance threshold. Threshold-crossing intervals that meet
        all other criteria are retained as place fields"""

        if tuning_curves is None:
            tuning_curves = self.tuning_curves
        if threshold_curves is None:
            threshold_curves = self.threshold_curves

        num_bins = tuning_curves.shape[1]
        min_width = num_bins * self.min_field_width
        max_width = num_bins * self.max_field_width

        place_fields = []
        for label, tuning in tuning_curves.iterrows():

            suprathres_positions = tuning > threshold_curves.loc[label]
            if all(suprathres_positions == 0):
                continue  # there are no suprathreshold events
            else:
                # find spatial intervals that exceed the null and evaluate
                # the other conditions to decide if it's a place field or not
                count = 0
                for start, stop in find_intervals(
                        suprathres_positions, circular=self.circular):

                    # apply conditions
                    field_width = stop - start if stop > start \
                            else stop + num_bins - start

                    if field_width < min_width:
                        continue
                    if field_width > max_width:
                        continue
                    if self.min_trials_active is not None:
                        if stop > start:
                            n_trials_active = self.rasters.loc[
                                    label].loc[:, start:(stop-1)].mean(
                                    axis=1).gt(0).sum()
                        else:
                            # handle the circular wrap-around
                            left_edge = self.rasters.loc[
                                    label, 0:(stop-1)].mean(axis=1).gt(0)
                            right_edge = self.rasters.loc[
                                    label, start:].mean(axis=1).gt(0)
                            n_trials_active = (left_edge + right_edge).sum()
                        if n_trials_active < self.min_trials_active:
                            continue

                    place_fields.append({'roi_label': label,
                                         'field_count': count,
                                         'start_bin': start,
                                         'stop_bin': stop})
                    count += 1

        self.place_fields = pd.DataFrame(place_fields, columns=[
                'roi_label', 'field_count', 'start_bin', 'stop_bin']
            ).set_index(['roi_label', 'field_count']).sort_index()

        return self.place_fields

    def _detect(self, signals, position, laps):
        """Detect place fields from the input signals. Along the way, this
        assigns the attributes self.tuning_curves, self.threshold_curves,
        self.null_curves, self.place_fields, and optionally, self.rasters."""
        # get observed tuning
        self.tuning_curves = self.tuning_strategy.calculate(
            signals, position, laps)
        if self.min_trials_active is not None:
            self.rasters = self.tuning_strategy.calculate_rasters(
                signals, position, laps)

        # generate null distribution and calculate thresholds
        self._get_nulls(signals, position, laps)
        self._get_thresholds()

        # detect place fields
        return self._find_fields()

    def _calculate_metrics(self):
        """doc string"""
        pass


class SignedPFDetector(SimplePFDetector):

    """Extension of SimplePFDetector that permits the detection of positive or
    negative ("anti") place fields, which can be useful for analyzing the
    activity of interneurons.

    """

    def _get_thresholds(self):
        """Basically we need to establish an upper and lower threshold
        at half the p-value on each side"""

        raise NotImplementedError

    def _find_fields(self):
        """"""

        pos_fields = super().find_fields(
            tuning_curves=self.tuning_curves,
            threshold_curves=self.pos_threshold_curves
        )
        pos_fields['sign'] = 'positive'

        neg_fields = super().find_fields(
            tuning_curves=self.tuning_curves * - 1,
            threshold_curves=self.neg_threshold_curves * -1
        )
        neg_fields['sign'] = 'negative'

        self.place_fields = pd.concat([pos_fields, neg_fields]).sort_index()

        return self.place_fields


class PeakPFDetector(PlaceFieldStrategy):

    """Detect place fields by finding cells with high peak-to-mean firing
    rate gain in the place field tuning curve, with some additional
    constraints. A simple detector that does not require shuffling.

    Parameters
    ----------
    peak_gain : float, optional
        Minimum peak-to-mean firing rate gain, computed from the spatial tuning
        curve, for each bin to qualify as being in a place field. Defaults to 3
    min_rate : float, optional
        Minimum mean activity rate. Defaults to 0 (no minimum)
    min_field_width, max_field_width : float, optional
        Minimum and maximum widths of place field, as fraction of belt length.
        Defaults to 0.03 and 0.5 respectively
    """

    def __init__(self, tuning_strategy, peak_gain=3, min_rate=0,
                 min_field_width=0.05, max_field_width=1):

        raise NotImplementedError



"""This submodule contains analysis routines for the analysis of spatial
tuning."""

from lab3.analysis.spatial.spatial_tuning import (
    SimpleSpatialTuning,
    CircularSimpleSpatialTuning
)

from lab3.analysis.spatial.place_fields import *



import json
from datetime import datetime
import numpy as np
import re
import os.path
import time
import pickle
from distutils.version import LooseVersion


# ------------------------------ MAIN FUNCTIONS ----------------------------- #


def find_files(directory, include_pickled=False):
    """Find tdml files in the specified directory

    Parameters
    ----------
    directory: str
        Path to search for tdml files
    include_pickled : bool, optional
        Whether to include a tdml file if a pickled version already exists at
        the same directory.

    Returns
    -------
    paths : list of str
        Paths for tdml files in directory
    """
    paths = []  # holds the paths
    pickled_paths = []  # temporally owns pickle files if they were found

    # checks if the path given isn't that of the tdml file
    if os.path.splitext(directory)[1] == '.tdml':
        paths.append(directory)

        # gets path of file without the .tdml
        path = os.path.splitext(directory)[0]

        # checks if there is .pkl file already present
        if os.path.exists('%s.pkl' % path):
            pickled_paths.append(path)  # stores the .pkl file

    # if we got a path for a directory
    else:
        # we walk through the directory and its sub-folders
        for dirpath, dirnames, filenames in os.walk(directory):
            # extends the paths array to hold all the .tdlm files' path
            paths.extend(
                map(lambda f: os.path.join(dirpath, f),
                    filter(lambda f: os.path.splitext(f)[1] == '.tdml',
                           filenames)))
            # extends the pickled_paths to hold all the .pkl file's path
            pickled_paths.extend(
                map(lambda f: os.path.join(dirpath, os.path.splitext(f)[0]),
                    filter(lambda f: os.path.splitext(f)[1] == '.pkl',
                           filenames)))

    if not include_pickled:
        # only return the paths that are not pickled
        paths = filter(
            lambda f: os.path.splitext(f)[0] not in pickled_paths, paths)

    return paths


def pickle_file(filepath):
    """Parses a tdml file and dumps information to pickle file

    Parameters
    ----------
    filepath : str
        Absolute or relative path to the file to parse
    """

    res = {}
    print(f'parsing {filepath}')

    last_ts = 0  # last time stamp

    with open(filepath, 'r') as f:
        for line in f:  # loops through lines
            try:
                data = json.loads(line)  # load json data of line
            except Exception:
                print(f'error parsing:\n {line}')

            last_ts = data.get('time', last_ts)  # gets the time stamp

            _time = data.get('time')
            parseKeys(
                globals(), data, _time, res, default, True,
                ['time', 'message'])

    if '__position_y' in res.keys():  # if position_y is in the key set
        res['__position_y'] = fixPosition(res['__position_y'],
                                          res['__position_updates'])

    if '__position_updates' in res.keys():
        res['__position_updates'] = \
                fillPositionTimes(res['__position_updates'])

    res['__trial_info']['laps'] = len(res.get('lap', []))

    # falling edge of the the 2nd sync pinc -- AOD bug: when acquiring
    # TSeries > 5min, acquisition has a 4s lag (vs ~500ms with <5min)
    if 'AOD_sync_pin' in res.keys():
        offset = res['AOD_sync_pin'][0][1]
        for key in filter(lambda k: k[0] != '_', res.keys()):
            if key not in ['trackLength', 'recordingDuration']:
                res[key] = res[key] - offset

    elif 'sync_pin' in res.keys():
        offset = res['sync_pin'][0][0]
        if np.isnan(offset):
            if np.isfinite(res['sync_pin'][0][1]):
                offset = res['sync_pin'][0][1] - 0.1
            else:
                offset = 0

        for key in filter(lambda k: k[0] != '_', res.keys()):
            if key not in ['trackLength', 'recordingDuration']:
                res[key] = res[key] - offset
    else:
        offset = 0

    if 'recordingDuration' not in res.keys():
        res['recordingDuration'] = last_ts - offset

    if "__position_y" in res.keys():
        res['__position_y'][:, 0] -= offset
        res['__position_y'] = res['__position_y'][np.where(
            res['__position_y'][:, 0] >= 0)]
        if res['__position_y'][0, 0] != 0:
            res['__position_y'] = np.insert(
                res['__position_y'], 0, [0, res['__position_y'][0, 1]], axis=0)
        treadmillPosition = np.array(res['__position_y'])
        treadmillPosition[:, 1] /= res['trackLength']
        treadmillPosition[:, 1] %= 1

        res['treadmillPosition'] = treadmillPosition

    if '__position_updates' in res.keys():
        res['__position_updates'][:, 0] -= offset
        res['__position_updates'] = res['__position_updates'][np.where(
            res['__position_updates'][:, 0] >= 0)]

    if 'stop_time' not in res['__trial_info'].keys():
        time_format = '%Y-%m-%d %H:%M:%S'
        start_ti = time.mktime(parseTime(res['__trial_info']['start_time']))
        res['__trial_info']['stop_time'] = datetime.fromtimestamp(
            start_ti + treadmillPosition[-1, 0]).strftime(time_format)

    output_file = f'{os.path.splitext(filepath)[0]}.pkl'
    print(f'writing {output_file}')
    with open(output_file, 'wb') as f:
        pickle.dump(res, f)

    print('finished writing')


def fix_NaNs_in_pkl(filename, save_result=True):
    """Attempts to fix NaNs present in the pickled data

    Parameters
    ----------
    filename : str
        Absolute or relative path to the file to fix
    save_results : bool, optional
        Flag that indicates whether to overwrite the pickled file
    """

    pkl_path = f'{os.path.splitext(filename)[0]}.pkl'
    with open(pkl_path, 'rb') as f:
        data = pickle.load(f)
    behavior_keys = filter(lambda k: re.match('^__.*', k) is None, data.keys())
    error_keys = [k for k in behavior_keys if np.any(np.isnan(data[k]))]
    context_info = data['__trial_info']['contexts']

    method = ''
    multiple_reward_flag = False
    for key in error_keys:
        cue_times = data[key]
        nan_times = np.where(np.isnan(cue_times))
        trial_length = data['__trial_info']['trial_length']
        track_length = float(data['__trial_info']['track_length'])

        if key in ['reward']:
            if 'reward' not in data['__reward_contexts']:
                multiple_reward_flag = True
                continue
            else:
                method = 'position'
        elif key in ['licking', 'water', 'tone']:
            method = 'time'
        elif re.match('.*_pin$', key) is not None:
            method = 'time'

        if key in context_info.keys():
            if context_info[key].get('locations') is not None:
                if isinstance(context_info[key]['locations'], list):
                    method = 'position'
                else:
                    method = 'position_shuffled'
            else:
                print('skipping %s' % key)
                continue
        print(f'{key} - {method}')
        if method == 'time':
            if key == 'sync_pin':
                cue_length = 0.1
            else:
                cue_length = np.nanmean(np.diff(cue_times))
                if cue_length == np.nan:
                    raise Exception("Unable to Determine Cue Length")
            for idx in nan_times[0][nan_times[1] == 1]:
                if cue_times[idx, 0] + cue_length < trial_length:
                    cue_times[idx, 1] = cue_times[idx, 0] + cue_length

            for idx in nan_times[0][nan_times[1] == 0]:
                if cue_times[idx, 1] - cue_length > 0:
                    cue_times[idx, 0] = cue_times[idx, 1] - cue_length

        elif method == 'position':
            position = data['treadmillPosition']
            radius = context_info[key]['radius'] / track_length
            max_duration = context_info[key].get('max_duration')
            for idx, j in zip(nan_times[0], nan_times[1]):
                if j == 1:
                    if max_duration is not None and \
                            cue_times[idx, 0] < max_duration:
                        continue
                    _position = position[position[:, 0] > cue_times[idx, 0]]
                    if max_duration is not None:
                        _position = _position[_position[:, 0]
                                              < _position[0, 0] + max_duration]
                else:
                    position_rev = position[::-1]
                    _position = position_rev[position_rev[:, 0]
                                             < cue_times[idx, 1]]
                    if max_duration is not None:
                        _position = _position[_position[:, 0]
                                              > _position[0, 0] - max_duration]

                if _position.shape[0] < 2:
                    continue

                location = None
                for location in context_info[key]['locations']:
                    location = location / track_length
                    if location - radius <= _position[0, 1] \
                            <= location + radius:
                        break
                if location is None:
                    raise Exception("start position not in context location")

                _position = _position[np.logical_not(np.logical_and(
                    location - radius <= _position[:, 1],
                    _position[:, 1] <= location + radius))]
                if _position.shape[0] == 0:
                    continue
                else:
                    cue_times[idx, j] = _position[0, 0]

        elif method == 'position_shuffled':
            position = data['treadmillPosition']
            radius = context_info[key]['radius'] / track_length
            max_duration = context_info[key].get('max_duration')
            for idx, j in zip(nan_times[0], nan_times[1]):
                if j == 1:
                    if max_duration is not None and \
                            cue_times[idx, 0] < max_duration:
                        continue
                    _position = position[position[:, 0] > cue_times[idx, 0]]
                    if max_duration is not None:
                        _position[_position[:, 0]
                                  < _position[0, 0] + max_duration]
                else:
                    position_rev = position[::-1]
                    _position = position_rev[position_rev[:, 0]
                                             < cue_times[idx, 1]]
                    if max_duration is not None:
                        _position[_position[:, 0]
                                  > _position[0, 0] - max_duration]

                if _position.shape[0] < 2:
                    continue

                if j == 1 and _position[1, 1] > _position[0, 1] or \
                        j == 0 and _position[1, 1] < _position[0, 1]:
                    location = _position[0, 1] + radius
                else:
                    location = _position[0, 1] - radius

                _position = _position[np.logical_not(np.logical_and(
                    location - radius <= _position[:, 1],
                    _position[:, 1] <= location + radius))]
                if _position.shape[0] == 0:
                    continue
                else:
                    cue_times[idx, j] = _position[0, 0]

    if multiple_reward_flag:
        all_reward_times = [data[k] for k in data['__reward_contexts']]
        all_reward_times = np.vstack(all_reward_times)
        data['reward'] = all_reward_times[all_reward_times[:, 0].argsort()]

    with open(pkl_path, 'r') as f:
        behavior_data = pickle.load(f)

    print('\nchanges to be made ...')
    for cue in error_keys:
        print(f'cue: {cue}')
        for idx in np.where(behavior_data[cue] != data[cue])[0]:
            print(f'{str(behavior_data[cue][idx])} -> {str(data[cue][idx])}')
        behavior_data[cue] = data[cue]

    if save_result:
        print('saving ...')
        with open(pkl_path, 'w') as f:
            pickle.dump(behavior_data, f)
    print('\n\n')


# -------------------------- TDML PARSING FUNCTIONS ------------------------- #


def parseTime(time_string):
    """Attempts to parse the given time string with an appropriate, readable
    format

    Parameters
    ----------
    time_string: str
        String to parse

    Returns
    -------
    time : parsed time
    """

    # acceptable time formats
    time_formats = ['%Y-%m-%d %H:%M:%S', '%d-%b-%Y %H:%M:%S',
                    '%Y-%m-%d%H:%M:%S', '%Y-%m-%d-%Hh%Mm%Ss']

    # loop through all the time formats
    for frmt in time_formats:
        try:  # attempt to parse the time, if successful return
            return time.strptime(time_string, frmt)
        except ValueError:  # otherwise pass
            pass


def parseKeys(functions, msg, _time, _dict, default=None, filtered=False,
              to_filter=None):
    """Attempts to call all the functions present in the message being passed.

    Parameters
    ----------
    :param functions: dictionary of functions present in the project
    :param msg: Json retrieved data (usally a dict)
    :param _time: time of data performed
    :param _dict: stores current trial's formatted info
    :param default: function to call if a key in message is not a previously
    defined function
    :param filtered: boolean that indicates if the message should be filtered
    :param to_filter: array of keys to filter
    :return: Void function
    """

    keys = msg.keys()  # get the keys from the message

    if filtered:  # check if we require some kind of filter
        if to_filter:  # check that we have the filter array
            # filter the list given the parameters
            keys = list(filter(lambda k: k not in to_filter, keys))

    for key in keys:  # loop through the keys
        # check if the function exists
        if '_{}_'.format(key) in functions.keys():
            # if it exists, call the function
            functions['_{}_'.format(key)](msg[key], _time, _dict)
        # else check that there is a default method to call
        elif default is not None:
            default(key, msg[key], _time, _dict)  # call default method


def default(key, value, _time, _dict):
    """
    Default method call for all parseKey processes
    :param key: if is to be edit
    :param value: if is to be edit
    :param _time: time of trial
    :param _dict: stores current trial's formatted info
    :type key: str
    :type value: str, int, bool, dict, etc ...
    :type _time: str
    :type _dict: dict
    :return: None
    :rtype: void
    """
    insert(_dict, '__trial_info', {str(key): value})


def insert(_dict, key, value):
    """
    Safe inserts a key in the dictionary
    :param _dict: stores current trial's formatted info
    :param key: to insert
    :param value: of the key
    :return: Void method
    """
    # attempt to get the key, if doesn't exists, it create an empty dictionary
    _dict[key] = _dict.get(key, {})
    _dict[key].update(value)  # update the key value with the intended value


def fillPositionTimes(position_updates):
    """
    populates missing times for the position_updates array
    :param position_updates: holds position time changes
    :type position_updates: np.array
    :return: updated position_updates
    :rtype: np.array
    """
    position_updates = np.flip(position_updates, axis=0)
    missingTimes = np.where([p is None for p in position_updates])[0]

    for t in missingTimes:
        position_updates[time, 0] = position_updates[t - 1, 0] \
                                    - position_updates[t, 2] * 10e-3
    return np.flip(position_updates, axis=0).astype(float)


def fixPosition(position_array, position_updates):
    """
    fixes the missing entries in the position_updates list
    :param position_array: position data
    :param position_updates: position time changes
    :type position_array: np.array
    :type position_updates: np.array
    :return: updated position data
    :rtype: np.array
    """
    position_updates = position_updates.astype(float)
    nans = np.where(np.isnan(position_updates))[0]
    for idx in np.flip(nans, 0):
        position_updates[idx + 1][1] += position_updates[idx][1]
        position_updates = np.delete(position_updates, idx, axis=0)

    delta = position_updates[1:, 0] - position_updates[1:, 2] * 10e-3
    new_times = delta[position_updates[:-1, 0] < delta]

    for i, t in enumerate(np.flip(new_times, axis=0)):
        idx = np.where(position_array[:, 0] > t)[0][0]
        position_array = np.insert(
            position_array, idx, [new_times[-(i + 1)], position_array[idx, 1]],
            axis=0)

    return position_array


def _version_(ver, _time, _dict):
    """
    checks trial's version comparability, and stores in information
    :param ver: version of current file
    :param _time: UNUSED
    :param _dict: stores current trial's formatted info
    :type ver: str
    :type _time: str
    :type _dict: dict
    :return: None
    :rtype: void
    """
    if LooseVersion(ver) < LooseVersion('0.0.6b5'):
        raise Exception('unsupported behaviorMate version')
    insert(_dict, '__trial_info', {'version': ver})


def _start_(time_str, _time, _dict):
    """
    adds the start time for the current trial to the reformatted data
    :param time_str: time to be parsed
    :param _time:  UNUSED
    :param _dict: stores current trial's formatted info
    :type time_str: str
    :type _time: str
    :type _dict: dict
    :return: None
    :rtype: void
    """
    sql_time_format = '%Y-%m-%d %H:%M:%S'
    insert(_dict, '__trial_info',
           {'start_time':
            time.strftime('%Y-%m-%d-%Hh%Mm%Ss',
                          time.strptime(time_str, sql_time_format))})


def _stop_(time_str, _time, _dict):
    """
    adds a stop time flag to the reformatted trial data
   :param time_str: time to be parsed
    :param _time:  UNUSED
    :param _dict: stores current trial's formatted info
    :type time_str: str
    :type _time: str
    :type _dict: dict
    :return: None
    :rtype: void
    """
    sql_time_format = '%Y-%m-%d %H:%M:%S'  # time format
    insert(_dict, '__trial_info',
           {'stop_time': time.strftime(
                    '%Y-%m-%d-%Hh%Mm%Ss',
                    time.strptime(time_str, sql_time_format))})
    if _dict.get('__trial_info', {}).get('trial_length') is None:
        insert(_dict, '__trial_info', {'trial_length': _time})


def _context_(msg, _time, _dict):
    """
    Ignores the context key word at a global level
    :param msg: ignored
    :param _time: ignored
    :param _dict: ignored
    :return:
    """
    pass


def _comments_(comments, _time, _dict):
    """
    Reformat comments for the current trial
    :param comments: comments to insert
    :param _time: UNUSED
    :param _dict: stores current trial's formatted info
    :type comments: str
    :type _time: str
    :type _dict: dict
    :return:
    """
    insert(_dict, '__trial_info', {'comments': comments})


def _settings_(msg, _time, _dict):
    """
    reformat settings information for the current trial
    :param msg:  dictionary with the trial specification
    :param _time: time of trial
    :param _dict: stores current trial's formatted info
    :type msg: dict
    :type _time: str
    :type _dict: dict
    :return: None
    :rtype: void
    """

    def _sync_pin_(msg, _time, _dict):  # adds the sync pin information
        insert(_dict, '__valve_ids', {msg: 'sync_pin'})

    def _track_length_(msg, _time, _dict):  # adds the track_length information
        _dict['trackLength'] = msg
        insert(_dict, '__trial_info', {'track_length': msg})

    def _trial_length_(msg, _time, _dict):  # adds the trial length information
        _dict['recordingDuration'] = msg
        insert(_dict, '__trial_info', {'trial_length': msg})

    # adds the general setting information for each context to the dictionary
    def _contexts_(msg, _time, _dict):
        contexts = {c['id']: c for c in msg}  # get's the name of each context
        for context in contexts.keys():  # traverse the whole list
            for valve in contexts[context].get('valves', []):
                # store the valve information separately
                insert(_dict, '__valve_ids', {valve: str('%s_pin' % context)})

        # add the context information to trial_info
        insert(_dict, '__trial_info', {'contexts': contexts})

    def _reward_(msg, _time, _dict):  # adds reward data to the file
        insert(_dict, '__trial_info', {'reward': msg})

    parseKeys(locals(), msg, _time, _dict, default=default)

    contexts = _dict['__trial_info'].get('contexts', {})

    reward_contexts = list(filter(lambda n: 'reward' in n, contexts.keys()))

    if 'reward' in _dict['__trial_info']:
        reward = _dict['__trial_info']['reward']
        if reward.get('id', 'reward') not in reward_contexts:
            reward_contexts.append(reward.get('id', 'reward'))

    reward_pins = []

    for context in reward_contexts:
        reward_pins.extend(contexts[context].get('valves'))

    if 'reward' in _dict['__trial_info']:
        if reward.get('pin') is not None and \
                reward.get('pin') not in reward_pins:
            reward_pins.append(reward['pin'])

    _dict['__reward_contexts'] = reward_contexts
    _dict['__reward_pins'] = reward_pins


def _behavior_mate_(msg, _time, _dict):
    """
    reformat messages posted by behavior_mate pc for the current trial
    :param msg:  dictionary with the position specifications
    :param _time: time of trial
    :param _dict: stores current trial's formatted info
    :type msg: dict
    :type _time: str
    :type _dict: dict
    :return: None
    :rtype: void
    """

    def _context_(msg, _time, _dict):
        if 'id' not in msg.keys():
            return

        _id = msg['id']
        if msg['action'] == 'start':
            if _dict.get(_id, np.zeros((1, 2)))[-1, 1] != np.NaN:
                _dict[_id] = np.vstack((
                    _dict.get(_id, np.empty((0, 2))), [_time, np.NaN]))

        elif msg['action'] == 'stop':
            if np.isnan(_dict.get(_id, np.zeros((1, 2)))[-1, 1]):
                _dict[_id][-1, 1] = _time

    if 'scene' in msg.keys():
        parseScene(msg, _dict)

    for key in msg.keys():
        if '_{}_'.format(key) in locals().keys():
            locals()['_{}_'.format(key)](msg[key], _time, _dict)


def _position_controller_(msg, _time, _dict):
    """
    reformat the position information for the current trial
    :param msg:  dictionary with the position specifications
    :param _time: time of trial
    :param _dict: stores current trial's formatted info
    :type msg: dict
    :type _time: str
    :type _dict: dict
    :return: None
    :rtype: void
    """

    def _lap_reset_(msg, _time, _dict, _millis):
        if 'position_lap_reader' not in _dict.keys():
            _dict['position_lap_reader'] = np.array([])

        _dict['position_lap_reader'] = np.append(
            _dict['position_lap_reader'], _time)

    def _position_(msg, _time, _dict, _millis):
        if '__position_updates' not in _dict.keys():
            _dict['__position_updates'] = np.array([]).reshape((0, 4))
        _dict['__position_updates'] = np.vstack(
            (_dict['__position_updates'], [_time, msg['dy'], msg['dt'],
                                           _millis]))

    for key in msg.keys():
        if '_{}_'.format(key) in locals().keys():
            locals()['_{}_'.format(key)](msg[key], _time, _dict,
                                         msg.get('millis'))


def _y_(msg, _time, _dict):
    """
    reformat the y component for the current trial
    :param msg:  dictionary with the trial specification
    :param _time: time of trial
    :param _dict: stores current trial's formatted info
    :type msg: dict
    :type _time: str
    :type _dict: dict
    :return: None
    :rtype: void
    """
    if '__position_y' not in _dict.keys():
        _dict['__position_y'] = np.array([]).reshape((0, 2))

    _dict['__position_y'] = np.vstack(
        (_dict['__position_y'], [_time, float(msg)]))


def _lap_(msg, _time, _dict):
    """
    reformat lap information
    :param msg:  dictionary with the lap specifications
    :param _time: time of trial
    :param _dict: stores current trial's formatted info
    :type msg: dict
    :type _time: str
    :type _dict: dict
    :return: None
    :rtype: void
    """
    if 'lap' not in _dict.keys():
        _dict['lap'] = np.array([])

    _dict['lap'] = np.append(_dict['lap'], _time)


def _behavior_controller_(msg, _time, _dict):
    """
    reformat the behavior information for the current trial
    :param msg:  dictionary with the position specifications
    :param _time: time of trial
    :param _dict: stores current trial's formatted info
    :type msg: dict
    :type _time: str
    :type _dict: dict
    :return: None
    :rtype: void
    """

    def context_update(this_id, this_msg, _time, _dict):
        if this_id not in _dict.keys():
            _dict[this_id] = np.array([]).reshape((0, 2))
        if this_msg.get('action') == 'start' or \
                this_msg.get('action') == 'open':
            _dict[this_id] = np.vstack((_dict[this_id], [_time, np.NaN]))
        elif this_msg.get('action') == 'stop' \
                or this_msg.get('action') == 'close':
            if _dict[this_id].shape[0] == 0 or not \
                    np.isnan(_dict[this_id][-1][-1]):
                _dict[this_id] = np.vstack(
                    (_dict[this_id], [np.NaN, _time]))
            else:
                _dict[this_id][-1][-1] = _time

    def _context_(msg, _time, _dict):
        context_id = str(msg['id'])  # get id from context
        if context_id in _dict.get('__valve_ids', {}).values() or \
                context_id == 'water':  #
            context_id = str(context_id) + '_'

        context_update(context_id, msg, _time, _dict)

        if context_id != 'reward' and \
                context_id in _dict.get('__reward_contexts'):
            msg['id'] = 'reward'
            _context_(msg, _time, _dict)

    def _valve_(msg, _time, _dict):
        valve_id = _dict.get('__valve_ids', {}).get(msg['pin'])

        if valve_id is None:
            valve_id = 'pin_%i' % msg['pin']

        context_update(valve_id, msg, _time, _dict)

        if msg['pin'] in _dict['__reward_pins']:
            context_update('water', msg, _time, _dict)

    def _tone_(msg, _time, _dict):
        valve_id = _dict.get('__tone_ids', {}).get(msg['pin'])

        if valve_id is None:
            valve_id = 'tone_%i' % msg['pin']

        context_update(valve_id, msg, _time, _dict)

    def _lick_(msg, _time, _dict):
        context_update('licking', msg, _time, _dict)

    parseKeys(locals(), msg, _time, _dict)


def parseScene(data, _dict):
    scene = data.get('id', data['scene'])
    if data['context']['action'] == 'start':
        if _dict.get(scene, np.zeros((1, 2)))[-1, 1] != np.NaN:
            _dict[scene] = np.vstack((
                _dict.get(scene, np.empty((0, 2))), [data['time'], np.NaN]))

    elif data['context']['action'] == 'stop':
        if np.isnan(_dict.get(scene, np.zeros((1, 2)))[-1, 1]):
            _dict[scene][-1, 1] = data['time']


# for populating the documentation website
__all__ = [
    'find_files',
    'pickle_file',
    'fix_NaNs_in_pkl'
]



import pandas as pd

from lab3.core.filters import Filter, _BoundFilter
from lab3.filters.time_filters import EventFilter

class TrialFilter(Filter):
    """Filter trials (e.g. PSTHs, lap-by-lap tuning curves) on some condition.

    Parameters
    ----------
    trial_name : str
        Trial identifier in the index of the full AnalysisResult dataframe. 
        This will usually be 'OnsetEvent', 'IntervalEvent', 'LapEvent', etc.
    trial_spec : dict
        Keyword arguments passed to expt.events(...) to get the trials
        of interest.
    """
    
    constructor_ids = Filter.constructor_ids + ['trial_name', 'trial_spec']

    def bind(self, trials):
        return _BoundFilter(index=trials, columns=None, 
                            level=self.trial_name,
                            inverse=self.inverse, 
                            name=self.name)

class RunningTrials(TrialFilter):
    def bind_to(self, expt): 
        try:
            as_time_events = self.trial_spec['as_time_events']
        except KeyError:
            as_time_events = False
        
        running_spec = {
            'signal_type': 'behavior',
            'event_type': 'running',
            'fmt': 'interval',
            'as_time_events': as_time_events
        }
        
        running = expt.events(**running_spec)
        
        trials = expt.events(**self.trial_spec)
        
        running_trials = pd.Index([i for i, trial in enumerate(trials) 
                                   if trial.onset in running])
        
        return self.bind(running_trials)



import sys
import re
import os
import argparse
from datetime import datetime
import matplotlib as mpl
# mpl.use('pdf')
import matplotlib.pyplot as plt
# from matplotlib.backends.backend_pdf import PdfPages

import numpy as np
import pandas as pd
import datetime
import pickle as pkl
import json
import numpy.ma as ma
from scipy.optimize import curve_fit
from scipy.interpolate import interp1d
# try:
#     from bottleneck import nanstd
# except ImportError:
#     from numpy import nanstd
from numpy import nanstd
# import cPickle as pickle
import itertools as it
import warnings

from lab3.experiment.base import BehaviorExperiment, ImagingExperiment, fetch_trials
from lab3.experiment.group import ExperimentGroup

from lab3.event.transient import Transient, SingleCellTransients, ExperimentTransientGroup
from lab3.filters import EventFilter

# setting up gobal variables for calculating transients
class configs:
    if (os.path.exists('configs.json')):
        settings = json.load(
            open('configs.json')).get('calculateTransients', {})
    else:
        settings = {}

    # keep first value .01 for initial pruning and baseline recalculation
    P_VALUES = settings.get('P_VALUES', [.01, .05])

    ##
    # Transient identification parameters
    ##
    ON_THRESHOLD = settings.get('ON_THRESHOLD', 2)  # threshold for starting an event (in sigmas)
    OFF_THRESHOLD = settings.get('OFF_THRESHOLD', 0.5)  # end of event (in sigmas)
    # for which to calculate parameters (all events greater are in final bin)
    MAX_SIGMA = settings.get('MAX_SIGMA', 5)
    MIN_DURATION = settings.get('MIN_DURATION', 0.1) # for which to calculate parameters
    MAX_DURATION = settings.get('MAX_DURATION', 5) # for which to calculate parameters
    N_BINS_PER_SIGMA = settings.get('N_BINS_PER_SIGMA', 2)
    N_BINS_PER_SEC = settings.get('N_BINS_PER_SEC', 4)
    # to iteratively identify events and re-estimate noise
    N_ITERATIONS = settings.get('N_ITERATIONS', 3)

    parameters = {
        'on_threshold': ON_THRESHOLD, 'off_threshold': OFF_THRESHOLD,
        'max_sigma': MAX_SIGMA, 'min_duration': MIN_DURATION,
        'max_duration': MAX_DURATION, 'n_bins_per_sigma': N_BINS_PER_SIGMA,
        'n_bins_per_sec': N_BINS_PER_SEC, 'n_iterations': N_ITERATIONS}

    parameters_text = 'ON_THRESHOLD = %.2f, OFF_THRESHOLD = %.2f, \
                       MAX_SIGMA = %.2f, MIN_DURATION = %.2f, \
                       MAX_DURATION = %.2f, N_BINS_PER_SIGMA = %d, \
                       N_BINS_PER_SEC = %d, N_ITERATIONS = %d' % \
        (float(ON_THRESHOLD), float(OFF_THRESHOLD), float(MAX_SIGMA),
         float(MIN_DURATION), float(MAX_DURATION), N_BINS_PER_SIGMA,
         N_BINS_PER_SEC, N_ITERATIONS)

    # one extra bin for events > the max time/sigma
    # parameters above must be chosen such that nBins are integers

    nTimeBins = int((MAX_DURATION - MIN_DURATION) * N_BINS_PER_SEC + 1)
    nSigmaBins = int((MAX_SIGMA - ON_THRESHOLD) * N_BINS_PER_SIGMA + 1)

def estimate_noise(expt, exclude_transients=False, channel='Ch2', label=None,
                   demixed=False):
    """
    Estimate the noise of each cell in the experiment.  If no transients have
    been identified, the noise is coarsely estimated to be the std of the trace
    (concatenated across cycles), a valid assumption for infrequently spiking
    pyramidal cells.  If transients are available, these epochs are excluded
    from the concatenated traces before calculating the std

    exclude_transients -- a transients structure (np record array)
    """
    
    signals = expt.signals(signal_type='dfof', label=label, channel=channel)
    imData = signals.to_numpy()[...,None]
    nCycles = imData.shape[2]
    # start_idx = expt.imagingIndex(10)
    # end_idx = expt.imagingIndex(45) - 1
    # imData = imData[:, start_idx:end_idx]
    concatenated_data = imData[:, :, 0]
    for i in range(nCycles - 1):
        concatenated_data = np.concatenate(
            (concatenated_data, imData[:, :, i + 1]), axis=1)

    if exclude_transients is True:

        expt_transients = expt.events(signal_type='imaging', event_type='transients', 
            label=label, channel=channel)

        activity = np.zeros(imData.shape, 'bool')

        if activity.ndim == 2:
            activity = activity.reshape(activity.shape[0],
                                        activity.shape[1], 1)

        for cell in expt_transients:
            roi_label  = cell.name
            cell_index = signals.level['roi_label'].get_loc(roi_label)

            for transient in cell:
                start = int(transient.onset)
                end = int(transient.offset)
                activity[cell_index, start:end + 1, 0] = True

        concatenated_activity = activity[:, :, 0]
        for i in range(nCycles - 1):
            concatenated_activity = np.concatenate(
                (concatenated_activity, activity[:, :, i + 1]), axis=1)

        # nan mask
        nan_mask = np.zeros(concatenated_activity.shape, dtype=bool)
        nan_mask[np.where(np.isnan(concatenated_data))] = True
        concatenated_activity = concatenated_activity | nan_mask

        masked_imData = ma.array(concatenated_data, mask=concatenated_activity)
        noise = masked_imData.std(axis=1).data

    else:
        noise = nanstd(concatenated_data, axis=1)


    return noise

def identify_events(data):
    # generator function which yields the start and stop frames for putative
    # events (data[start, stop] yields the full event)

    # accepts data in terms of sigmas
    L = len(data)
    start_index = 0

    while start_index < L:
        starts = np.where(data[start_index:] > configs.ON_THRESHOLD)[0].tolist()
        if starts:
            # start is the frame right before it crosses ON_THRESHOLD
            # (inclusive)
            abs_start = np.max([0, start_index + starts[0] - 1])
        else:
            break

        ends = np.where(data[abs_start + 1:] < configs.OFF_THRESHOLD)[0].tolist()
        if ends:
            # end is the first frame after it crosses OFF_THRESHOLD (event is
            # inclusive of abs_end -- need to add 1 to include this frame when
            # slicing)
            abs_end = abs_start + ends[0]
        else:
            break

        start_index = abs_end + 1

        yield abs_start, abs_end
#         wrap this with event constructor. lab3.events
#         get rid of yield and return an event group instead

def add_events_to_histogram(data, sigma, frame_period, direction, counter):
    """
    Given a timeseries (data), identify putative events and add them to counter
    """

    if direction is 'negative':
        data = -1 * data / sigma
    else:
        data = data / sigma

    for abs_start, abs_end in identify_events(data):

        dur = (abs_end - abs_start) * frame_period
        amp = np.nanmax(data[abs_start:abs_end + 1])

        if np.isnan(amp):
            continue

        if dur >= configs.MIN_DURATION:
            # bins are of the form [start, end)
            # add one for slicing
            sigma_bin_ind = int(np.floor(configs.N_BINS_PER_SIGMA * amp) -
                                configs.ON_THRESHOLD * configs.N_BINS_PER_SIGMA + 1)
            if sigma_bin_ind > configs.nSigmaBins:
                sigma_bin_ind = configs.nSigmaBins
            # add one for slicing
            time_bin_ind = int(np.floor(configs.N_BINS_PER_SEC * dur) -
                               configs.MIN_DURATION * configs.N_BINS_PER_SEC + 1)
            if time_bin_ind > configs.nTimeBins:
                time_bin_ind = configs.nTimeBins

            counter[:sigma_bin_ind, :
                    time_bin_ind] += np.ones([sigma_bin_ind, time_bin_ind])

    return counter

def calculate_event_histograms(experimentList, exclude_transients=False,
                               channel='Ch2', label=None, demixed=False):
    """
    Recursively calls add_events_to_histogram in order to pool event across
    cells across experiments
    """

    negative_event_counter = np.zeros((configs.nSigmaBins, configs.nTimeBins))
    positive_event_counter = np.zeros((configs.nSigmaBins, configs.nTimeBins))
    noise_dict = {}

    for expt in experimentList:
        # if exclude_transients is None:
        #     exclusion = None
        # else:
        #     exclusion = exclude_transients[expt]

#         this method is not implemented in lab3 so i decided to comment it out.

        # valid_filter = expt.validROIs(
        #     fraction_isnans_threshold=0, contiguous_isnans_threshold=0,
        #     dFOverF='from_file', channel=channel, label=label, demixed=demixed)
        valid_filter = None

        frame_period = expt.frame_period

        signals = expt.signals(signal_type='dfof', label=label, channel=channel)
        imData = signals.to_numpy()[...,None]

        noise_dict[expt.trial_id] = estimate_noise(expt, exclude_transients=exclude_transients,
                                          channel=channel, label=label,
                                          demixed=demixed)

#         valid_indices = expt._filter_indices(
#             valid_filter, channel=channel, label=label)

        for cell_data, sigma in zip(imData, noise_dict[expt.trial_id]): #[valid_indices]):
            for cycle_data in cell_data.T:
                positive_event_counter = add_events_to_histogram(
                    cycle_data, sigma, frame_period, 'positive',
                    positive_event_counter)
                negative_event_counter = add_events_to_histogram(
                    cycle_data, sigma, frame_period, 'negative',
                    negative_event_counter)

    return positive_event_counter, negative_event_counter, noise_dict


def calculate_transient_thresholds(experimentList, p=[.05], return_figs=False,
                                   fit_type='pw_linear',
                                   exclude_transients=False, channel='Ch2',
                                   label=None, demixed=False):

    if fit_type is None:
        fit_type = 'pw_linear'

    positive_events, negative_events, noise = calculate_event_histograms(
        experimentList, exclude_transients=exclude_transients, channel=channel,
        label=label, demixed=demixed)

    fpr = negative_events / positive_events

    d = np.linspace(configs.MIN_DURATION, configs.MAX_DURATION, configs.nTimeBins)
    d_fit_axis = np.linspace(configs.MIN_DURATION, configs.MAX_DURATION, 10000)
    thresholds = np.empty([configs.nSigmaBins, len(p)])

    if return_figs:
        nRows = 2
        nCols = 2
        nHistFigs = int(np.ceil(configs.nSigmaBins / float(nRows * nCols)))

        figs = [[] for x in range(nHistFigs)]
        axs = []
        fpr_fig, fpr_ax = plt.subplots(1, 1, figsize=(15, 8))

        for fig in xrange(nHistFigs):
            figs[fig], ax = plt.subplots(nRows, nCols, figsize=(15, 8))
            if ax.ndim == 1:
                ax = ax[:, np.newaxis]
            axs = np.hstack([axs, ax.flatten()])

        colors = plotting.color_cycle()
        color_list = []
        for sigma_ind in range(configs.nSigmaBins):
            ax = axs[sigma_ind]
            ax.bar(d, positive_events[sigma_ind],
                   width=1. / configs.N_BINS_PER_SEC, color='b')

            ax.bar(-1 * d - 1. / configs.N_BINS_PER_SEC,
                   negative_events[sigma_ind], width=1. / configs.N_BINS_PER_SEC,
                   color='r')

            ax.set_xticks(np.arange(-configs.MAX_DURATION, configs.MAX_DURATION + 1, 1))
            ax.set_xticklabels(np.arange(-configs.MAX_DURATION, configs.MAX_DURATION + 1, 1))

            ax.set_xlabel('Duration (sec)')
            ax.set_ylabel('Number of events')
            sigma_level = np.around(
                configs.ON_THRESHOLD + sigma_ind * (1. / configs.N_BINS_PER_SIGMA), decimals=3)
            ax.set_title('>= %s Sigma' % str(sigma_level))

            c = colors.next()
            fpr_ax.plot(d, fpr[sigma_ind], color=c,
                        marker='o', markersize=4, ls='')
            color_list.append(c)

    def exp_fit_func(x, a, b, c):
        return a * np.expt(-b * x) + c

    labels_list = [[] for x in range(configs.nSigmaBins)]
    for sigma_ind, sigma_fpr in enumerate(fpr):
        # pull out the monotonically decreasing part of the fpr curves --
        # currently not used for anything
        monotonic_inds = np.where(np.diff(sigma_fpr) > 0)[0] + 1
        if len(monotonic_inds) > 0:
            d1 = d[:monotonic_inds[0]]
            sigma_fpr1 = sigma_fpr[:monotonic_inds[0]]
            for fpr_ind in range(monotonic_inds[0], len(sigma_fpr)):
                if sigma_fpr[fpr_ind] < sigma_fpr1[-1]:
                    d1 = np.append(d1, d[fpr_ind])
                    sigma_fpr1 = np.append(sigma_fpr1, sigma_fpr[fpr_ind])
        else:
            # all monotonic
            d1 = d
            sigma_fpr1 = sigma_fpr

        sigma_level = np.around(
            configs.ON_THRESHOLD + sigma_ind * (1. / configs.N_BINS_PER_SIGMA), decimals=3)
        labels_list[sigma_ind] = '%s Sigma -- ' % (str(sigma_level))

        fit = None  # curve to plot if interpolation is performed
        for p_ind, p_val in enumerate(p):

            if sigma_fpr[0] < p_val:
                threshold = 0
            elif len(np.where(sigma_fpr < p_val)[0]) == 0:
                threshold = np.nan
            else:
                if fit_type is 'exponential':
                    popt, pcov = curve_fit(exp_fit_func, d, sigma_fpr)
                    fit = exp_fit_func(d_fit_axis, popt[0], popt[1], popt[2])
                    threshold = -1 * np.log(
                        (p_val - popt[2]) / popt[0]) / popt[1]
                    if threshold < 0:
                        threshold = 0

                elif fit_type is 'polynomial':
                    z = np.polyfit(d, sigma_fpr, 4)
                    f = np.poly1d(z)
                    fit = f(d_fit_axis)

                    roots = (f - p_val).r
                    real_positive_roots = roots[
                        np.isreal(roots) * roots.real > 0]
                    if len(real_positive_roots) > 0:
                        threshold = np.amin(real_positive_roots)
                    else:
                        threshold = np.nan

                elif fit_type is 'pw_linear':
                    f = interp1d(sigma_fpr1[::-1], d1[::-1],
                                 kind='linear', bounds_error=False)

                    d_fit_axis = d1
                    fit = sigma_fpr1

                    threshold = f(p_val)

            thresholds[sigma_ind, p_ind] = threshold

            threshold_str = str(
                np.around(thresholds[sigma_ind, p_ind], decimals=3))
            if p_ind == len(p) - 1:
                labels_list[
                    sigma_ind] += 'p=%s: %s' % (str(p_val), threshold_str)
            else:
                labels_list[
                    sigma_ind] += 'p=%s: %s, ' % (str(p_val), threshold_str)

        if return_figs:
            if fit is not None:
                fpr_ax.plot(d_fit_axis, fit, color=color_list[sigma_ind])

    if return_figs:
        fpr_ax.set_ylim((-.05, 0.8))
        fpr_ax.set_xlabel('Duration (sec)')
        fpr_ax.set_ylabel('False positive rate (p)')
        fpr_ax.set_title('False positive rates by sigma and duration')
        plotting.stackedText(fpr_ax, textList=labels_list,
                             colors=color_list, loc=1)

        figs = np.hstack([figs, fpr_fig])
        return thresholds, noise, figs
    else:
        return thresholds, noise

def identify_transients(experimentList, thresholds, noise=None, channel='Ch2',
                        label=None, demixed=False, save=False):

    transients = {}
    for expt in experimentList:
        signals = expt.signals(signal_type='dfof', label=label, channel=channel)
        imData = signals.to_numpy()[...,None]
        
        (nCells, _, nCycles) = imData.shape

        if noise is None:
            exp_noise = estimate_noise(expt, exclude_transients=False)
        else:
            exp_noise = noise[expt.trial_id]

        frame_period = expt.frame_period
        
        expt_transients = []
        
        for cell_idx, roi_label, cell_data, sigma in zip(
                it.count(), signals.index, imData, exp_noise):
            
            cell_transients = [] #SingleCellTransients([], name=roi_label)

            for cycle_idx, cycle_data in zip(it.count(), cell_data.T):
                for start, stop in identify_events(cycle_data / sigma):
                    amp = np.nanmax(cycle_data[start:stop + 1])
                    if np.isnan(amp):
                        continue
                    dur = (stop - start) * frame_period

                    sigma_bin_ind = int(np.floor(configs.N_BINS_PER_SIGMA *
                                        (amp / sigma)) - configs.ON_THRESHOLD *
                                        configs.N_BINS_PER_SIGMA)
                    if sigma_bin_ind > configs.nSigmaBins - 1:
                        sigma_bin_ind = configs.nSigmaBins - 1

                    if dur > thresholds[sigma_bin_ind] and \
                            dur > configs.MIN_DURATION:
                        
                        rel_max_ind = np.where(
                            cycle_data[start:stop + 1] == amp)[0].tolist()[0]
                        transient = Transient(onset=start, offset=stop, sigma=sigma, 
                                          duration=dur, amplitude=amp, 
                                          amplitude_index=start + rel_max_ind)
                        
                        cell_transients.append(transient)

                cell_transients = SingleCellTransients(cell_transients, name=roi_label)
                
            if len(cell_transients) > 0:
                expt_transients.append(cell_transients)
            

        expt_transients = ExperimentTransientGroup(expt_transients, name=expt.trial_id)
            
        if save:
            expt_transients.save(expt.imaging_events_path, event_type='transients', 
                                 label=label, channel=channel)

        transients[expt.trial_id] = expt_transients
        
    return transients



import numpy as np
import os 

def check_kwds(path, kwds):
    if kwds is None:
        return True
    else:
        word_in_path = [word in path for word in kwds]
        return np.all(word_in_path)

def get_sima_paths(paths, keywords=None):
    sima_paths = []
    for path in paths:
        for dirpath, _, _ in os.walk(path):
            if dirpath.endswith('.sima') and check_kwds(dirpath, keywords):
                sima_paths.append(dirpath)

    return sorted(sima_paths)


import sima
import pickle

from os.path import join

COMPATIBLE_PROTOCOL = 2
NATIVE_SEQUENCES = set(["Sequence",
                    "_AddWrapperSequence",
                    "_DivideWrapperSequence",
                    "_FloorDivideWrapperSequence",
                    "_IndexedSequence",
                    "_Joined_Sequence",
                    "_MaskedSequence",
                    "_MathWrapperSequence",
                    "_MotionCorrectedSequence",
                    "_MultiplyWrapperSequence",
                    "_Sequence_HDF5",
                    "_Sequence_TIFF_Interleaved",
                    "_Sequence_TIFFs",
                    "_Sequence_constant",
                    "_Sequence_memmap",
                    "_Sequence_ndarray",
                    "_SubtractWrapperSequence",
                    "_TrueDivideWrapperSequence",
                    "_WrapperSequence"])

class SIMACompatibilityMixin:
    def save(self, savedir=None):
        """Save the ImagingDataset to a file."""
        print("Saving in SIMA compatible format...")
        if savedir is None:
            savedir = self.savedir
        self.savedir = savedir

        if self._read_only:
            raise Exception('Cannot save read-only dataset.  Change savedir ' +
                            'to a new directory')
        # Keep this out side the with statement
        # If sequences haven't been loaded yet, need to read sequences.pkl
        
        # Saves as the best-match parent class AVAILABLE FROM VANILLA SIMA
        sequences = []
        for seq in self.sequences:
            seq_dict = seq._todict(savedir)
            native_sima_sequence = get_native_sima_sequence(seq)
            seq_dict['__class__'] = native_sima_sequence
            sequences.append(seq_dict)

        with open(join(savedir, 'sequences.pkl'), 'wb') as f:
            pickle.dump(sequences, f, COMPATIBLE_PROTOCOL)

        with open(join(savedir, 'dataset.pkl'), 'wb') as f:
            pickle.dump(self._todict(), f, COMPATIBLE_PROTOCOL)

def get_native_sima_sequence(custom_sequence):
    """Finds the best match parent sequence which is available
    from vanilla SIMA. 
    """
    seq = custom_sequence.__class__
    while not seq.__name__ in NATIVE_SEQUENCES:
        seq = seq.__base__

    if seq is object:
        raise TypeError("No native sequence found! (are you sure this is a SIMA sequence?)")
    else:
        return seq

def sima_compatible(klass):
    if issubclass(klass, sima.ImagingDataset):
        klass.save = SIMACompatibilityMixin.save
    elif issubclass(klass, sima.Sequence):
        def custom_todict(self, savedir=None):
            d = self._old_todict(savedir)
            d['__class__'] = get_native_sima_sequence(self)
            return d

        klass._old_todict = klass._todict
        klass._todict = custom_todict
    else:
        raise TypeError(f"Not sure how to handle {klass.__name__}. " + \
                f"Is this a Sequence or ImagingDataset?")
    
    return klass




"""Tools for handling signal extraction and import via SIMA and Suite2p"""

from lab3.extraction.s2p.base import Suite2pImagingDataset
from lab3.extraction.s2p.batch import BatchSuite2pImagingDataset



import sima
import read_roi
import numpy as np

def imagej2sima(roi_zip_path, im_shape=None):
    rois = []
    rlist = read_roi.read_roi_zip(roi_zip_path)
    for label, imagej_roi in rlist.items():
        x_coords = imagej_roi['x']
        y_coords = imagej_roi['y']
        z = imagej_roi['position']

        n_vertices = len(x_coords)

        polygon = np.zeros((n_vertices, 3))
        polygon[:, 0] = x_coords
        polygon[:, 1] = y_coords
        polygon[:, 2] = z - 1

        roi = sima.ROI.ROI(polygons=polygon, label=label, id=label, im_shape=im_shape)
        rois.append(roi)

    print("{} ROIs loaded from ImageJ.".format(len(rlist)))

    return sima.ROI.ROIList(rois)



import os

import numpy as np
import pandas as pd
from scipy.sparse import coo_matrix
from sima.ROI import ROIList, ROI

from lab3.signal import SignalFile
from lab3.misc.progressbar import ProgressBar


def import_to_signals_file(ds, label='suite2p', channel='Ch2',
                           overwrite=False):
    """Import Suite2p signal results using the new-style signal formats
    (dataframes and h5 stores).

    Parameters
    ----------
    ds : Suite2pStrategy
        Instance of a Suite2pStrategy, which has been previously extracted.
    label : str, optional
        ROI label to store the signals and ROI masks under. Defaults to
        'suite2p'.
    channel : {str or int}, optional
        Name or index of dynamic channel in the imaging dataset. Defaults to
        'Ch2'.
    overwrite : bool, optional
        If key '/{channel}/{label}' already exists in the dataset signals file,
        overwrite with new import. Note this will delete any additional signal
        types in the store in addition to the 'raw' and 'npil' (e.g. 'dfof'),
        to prevent mixtures of signals derived from different imports. Defaults
        to False, which will raise an error if the desired key already exists.
    """

    # check if key already exists in signals file, and deal with overwrites
    with SignalFile(os.path.join(ds.savedir, "signals.h5")) as signal_file:
        base_key = f"/{channel}/{label}"
        if base_key in signal_file:
            assert overwrite, f"Signals already exist for {base_key}. " \
                + "Set overwrite=True to re-import.\nCurrent file structure:" \
                + f" \n {signal_file.info()}"

            print(f"Deleting previous import at {base_key}")
            signal_file.remove(base_key)

        # create a sima ROI list and save to the underlying imaging dataset
        print("Making rois...")
        rois = _make_roi_masks(ds)
        print(f"Made {len(rois)} rois")
        ds.add_ROIs(rois, label=label)

        # create signal dataframes
        # TODO replace this with proper signal classes for each type,
        # and store metadata
        idx = pd.Index([roi.label for roi in rois], name='roi_label')
        raw = pd.DataFrame(ds.raw_signals[ds.cell_indicator], index=idx)
        npil = pd.DataFrame(ds.npil_signals[ds.cell_indicator], index=idx)

        try:
            raw.loc[:, ds.bad_frames] = np.nan
            npil.loc[:, ds.bad_frames] = np.nan
            print(f"Masked frames {ds.bad_frames}")
        except FileNotFoundError:
            pass

        # Handle a suite2p quirk - 0 means 'nan'
        raw[raw == 0] = np.nan
        npil[npil == 0] = np.nan

        # store signals
        # timestamp = datetime.strftime(datetime.now(), '%Y-%m-%d-%Hh%Mm%Ss')
        print("Storing...")
        signal_file.put(f"{base_key}/raw", raw)
        signal_file.put(f"{base_key}/npil", npil)
        print(signal_file.info())


def _make_roi_masks(ds):
    """Parse Suite2p results files to create an ROI object for each cell

    Parameters
    ----------
    ds : Suite2pStrategy
        Instance of a Suite2pStragegy, which has been previously extracted.

    Returns
    -------
    rois : sima.ROI.ROIList
    """

    num_planes, num_y, num_x = ds.sequences[0].shape[1:-1]

    # get row index of identified cells. note this is the position of each cell
    # in the total matrix of signals (combined across planes)
    cell_idx = ds.cell_indicator

    rois = [] * len(cell_idx)
    cum_rois = 0
    blank_mask = [coo_matrix((num_y, num_x)) for z in range(num_planes)]
    claimed_identifiers = []
    for z in range(len(ds.ops_list)):
        # get ROI information for current plane
        print(f"Plane {z}")
        stat = ds.get_stat_file(plane=z)

        # get the subset of cell_idx that appear in the current plane,
        # and shift the indices to match those in the plane stat file
        plane_idx = (cell_idx >= cum_rois) \
            & (cell_idx < (cum_rois + len(stat)))
        plane_idx = cell_idx[plane_idx] - cum_rois

        # get info for each valid ROI from stat and create mask
        p = ProgressBar(len(stat[plane_idx]))
        for i, sdict in enumerate(stat[plane_idx]):
            mask = list(blank_mask)
            mask[z] = coo_matrix((sdict['lam'], (sdict['ypix'],
                                  sdict['xpix'])),
                                 shape=(num_y, num_x))
            identifier = '{0:04d}-{1:04d}-{2:04d}'.format(
                z, int(np.mean(sdict['ypix'])),
                int(np.mean(sdict['xpix'])))
            if identifier in claimed_identifiers:
                # augment identifier with a number to denote the nth ROI that
                # shares the same centroid
                num_shared_centroids = np.sum(
                    [identifier in i for i in claimed_identifiers])
                identifier += f'_{num_shared_centroids}'

            # add identifier to claimed list
            claimed_identifiers.append(identifier)

            roi = ROI(mask=mask, label=identifier, id=identifier)
            rois.append(roi)
            p.update(i)
        print(' Done!')
        cum_rois += len(stat)

    return ROIList(rois)


def import_to_pkl():
    """For old-style imports"""
    raise NotImplementedError



from abc import abstractmethod

import pandas as pd

from lab3.core import Automorphism


class DecompositionStrategy(Automorphism):

    """Base class for signal decomposition strategies"""

    name = "Decomposition"

    def apply_to(self, experiment, channel='Ch2', label=None, from_type='dfof',
                 **kwargs):
        """Apply decomposition analysis to an experiment object.

        Parameters
        ----------
        experiment : lab3.experiment.base.ImagingExperiment
            Instance of ImagingExperiment to analyze
        channel : str or int, optional
        label : str, optional
        from_type : str {'dfof', 'raw'} or custom, optional

        Additional keyword arguments are passed directly to `calculate`.
        """
        sig_dict = self._load_signals(experiment, channel=channel, label=label,
                                      from_type='dfof')
        return self.calculate(**sig_dict, **kwargs)

    def calculate(self, signals):
        """Do analysis"""

        return self._decompose(signals)

    # ----------------- Implement these methods to subclass ----------------- #

    @abstractmethod
    def _decompose(self, signals):
        """Implement the decomposition algorithm here. It should return a
        dataframe the same width as signals, preserving the column index
        """

    # ------------------------------- Optional ------------------------------ #

    def _load_signals(self, experiment, channel, label, from_type):
        """By default only a single signal dataframe is passed to `calculate`
        when using `apply`. If you override `calculate` to take additional
        arguments in your custom strategy, you should load and add them to the
        dictionary here, which will be unpacked as arguments to `calculate`.
        """
        sig_dict = {
            'signals': experiment.signals(signal_type=from_type,
                                          channel=channel, label=label,
                                          max_frame=None)
        }
        return sig_dict


class LinearDynamicalSystem(DecompositionStrategy):

    """Fit a linear dynamical system using SSM.

    Parameters
    ----------
    n_components : int, optional
        Dimensionality of the latent space. Defaults to 1

    Additional keyword arguments are passed directly to ssm.LDS.fit during
    calculation.
    """

    def __init__(self, n_components=1, **model_kws):

        self.n_components = 1
        self.model_kws = model_kws

    def _decompose(self, signals):

        from ssm import LDS
        self.model = LDS(signals.shape[0], self.n_components, **self.model_kws)
        self.model_outputs = self.model.fit(signals.values.T)
        latents = pd.DataFrame(
            self.model_outputs[1].mean_continuous_states[0].T)
        latents.columns = signals.columns
        latents.index.name = 'component'

        return latents
